"""Distributional plots."""
from itertools import cycle

from bokeh.models import HoverTool
import pandas as pd
from seaborn._statistics import Histogram

from ezcharts.plots import BokehPlot, util


__all__ = ["displot", "histplot", "kdeplot", "ecdfplot", "rugplot", "distplot"]


def displot(*args, **kwargs):
    """Figure-level interface for distribution plots in a grid."""
    raise NotImplementedError


def histplot(
    data=None, *, x=None, y=None, hue=None, weights=None,
    stat='count', bins='auto', binwidth=None, binrange=None,
    discrete=None, cumulative=False, common_bins=True,
    common_norm=True, multiple='layer', element='bars',
    fill=True, shrink=1, kde=False, kde_kws=None,
    line_kws=None, thresh=0, pthresh=None, pmax=None,
    cbar=False, cbar_ax=None, cbar_kws=None, palette=None,
    hue_order=None, hue_norm=None, color=None, log_scale=None,
        legend=True, ax=None, **quad_kwargs):
    """Plot univariate or multivariate histograms."""
    plt = BokehPlot()

    estimate_kws = dict(
        stat=stat,
        bins=bins,
        binwidth=binwidth,
        binrange=binrange,
        discrete=discrete,
        cumulative=cumulative,
    )
    data = pd.DataFrame(data)

    estimator = Histogram(**estimate_kws)

    if data.ndim > 1 and data.shape[1] > 1:
        # multivariate data
        opacity = 0.5
        if palette is None:
            palette = util.choose_palette()
    else:
        opacity = 1.0
        if color is None:
            palette = util.choose_palette()
        else:
            palette = [color]
    if hue:
        data = data.pivot(columns=hue, values=data.columns[0])
    # this just looks over values if data is 1D
    # for var, color in zip(data, cycle(palette)):
    for col, color in zip(data.columns, cycle(palette)):
        quad_kwargs = {}
        if len(data.columns) > 1:
            quad_kwargs["legend_label"] = col
        variable_data = data[col].dropna()
        heights, edges = estimator(variable_data, weights=weights)

        plt._fig.quad(
            top=heights, bottom=0, left=edges[:-1], right=edges[1:],
            fill_color=color, fill_alpha=opacity, line_color=color, **quad_kwargs
        )
    plt._fig.y_range.start = 0
    hover = plt._fig.select(dict(type=HoverTool))
    hover.tooltips = [(stat.capitalize(), "@top")]
    return plt


def kdeplot(*args, **kwargs):
    """Plot uni/bi-variate distributions using kernel density estimation."""
    raise NotImplementedError


def ecdfplot(*args, **kwargs):
    """Plot empirical cumulative distribution functions."""
    raise NotImplementedError


def rugplot(*args, **kwargs):
    """Plot marginal distributions by drawing ticks along the x and y axes."""
    raise NotImplementedError


def distplot(*args, **kwargs):
    """Flexibly plot a univariate distribution of observations (DEPRECATED)."""
    raise DeprecationWarning
    raise NotImplementedError



"""Visualizing plasmids."""
import json
import os


from Bio import SeqIO
from dominate.tags import button, div, h6, input_, main, option, script, select
from dominate.util import raw

from ezcharts.layout.resource import VendorResource


__all__ = ["seqviz"]


def seqviz(plannotate_json, fasta, alias):
    """Visualize a plasmid using SeqViz."""
    if not os.path.exists(fasta):
        raise FileNotFoundError(f"{fasta} not found")

    if not os.path.exists(plannotate_json):
        raise FileNotFoundError(f"{plannotate_json} not found")

    records = list(SeqIO.parse(fasta, "fasta"))
    if len(records) != 1:
        raise ValueError(f"Expected 1 record in {fasta}, found {len(records)}")

    sequence = records[0].seq
    plannotate = json.load(open(plannotate_json))

    # features
    features = [
        dict(
            start=feature["Start Location"],
            end=feature["End Location"],
            name=feature["Feature"],
            direction=feature["Strand"]
         ) for feature in plannotate[alias]["features"]
    ]

    # primers { start: 33, end: 53, name: "LacZ Foward Primer", direction: 1 }
    primers = list()
    if 'primers' in plannotate[alias]:
        primers = [
            dict(
                start=primer["Start Location"],
                end=primer["End Location"],
                name=primer["Feature"],
                direction=primer["Strand"],
                color="#FAA887"
            ) for primer in plannotate[alias]["primers"]
        ]

    # place holder until we get the restriction enzymes from the plannotate json
    enzymes = list()
    if 'enzymes' in plannotate[alias]:
        enzymes = [enzyme['name'] for enzyme in plannotate[alias]["enzymes"]]

    # add the seqviz js
    seqviz_js = VendorResource("seqviz.min.js", tag=script)
    seqviz_js()

    with div(className="container"):
        with div(id="seqviz", cls="bg-light border p-4"):
            button(
                "Toggle controls",
                cls="btn btn-primary",
                type="button",
                onclick="showDiv()")

            with div(
                    cls="border p-4 my-3 bg-white",
                    style="display:none;",
                    id="controls-seqviz"):
                h6("Search")
                input_(
                    placeholder="Search for a sequence...",
                    type="text",
                    cls="form-control mb-3",
                    value="",
                    id="seqviz_search")

                if len(enzymes) > 1:
                    h6("Restriction enzymes")

                    # add search for multi-select box
                    input_(
                        placeholder="Search for an enzyme...",
                        type="text",
                        cls="form-control mb-2",
                        id="enzyme_search",
                        onkeyup="filterSelect()")

                    # add the multi-select box
                    with select(
                        cls="form-select mb-3",
                        multiple=True,
                        id="enzyme_select",
                        onchange="seqviz_render()"
                    ):
                        for enzyme in enzymes:
                            option(enzyme, value=enzyme)

                    button(
                        "Select all",
                        type="button",
                        cls="btn btn-primary btn-sm",
                        onclick="toggleSelectAll(true)")

                    button(
                        "Select none",
                        type="button",
                        cls="btn btn-primary btn-sm",
                        onclick="toggleSelectAll(false)")

            div(id="seqviz_container")

        with main(className="seqviz_plasmid"):
            div(
                raw(f"""
                    <script>
                        seqviz_render()
                        document.getElementById(
                            "seqviz_search").addEventListener("keyup", seqviz_render);

                        // show our controls div
                        function showDiv() {{
                            var div = document.getElementById('controls-seqviz');
                                if (div.style.display !== 'none') {{
                                    div.style.display = 'none';
                                }}
                                else {{
                                    div.style.display = 'block';
                                }}
                        }};

                        // toggle select all or none for the multi-select
                        function toggleSelectAll(state) {{
                            var select = document.getElementById('enzyme_select');
                            for (var i = 0; i < select.options.length; i++) {{
                                select.options[i].selected = state;
                                }}
                            seqviz_render();
                        }}

                        // filter the multi-select dropdown
                        function filterSelect() {{
                            var input = document.getElementById('enzyme_search');
                            var filter = input.value.toLowerCase();
                            var select = document.getElementById('enzyme_select');
                            var options = select.options;

                            for (var i = 0; i < options.length; i++) {{
                                var txtValue = options[i].textContent
                                    || options[i].innerText;
                                options[i].style.display = (
                                    txtValue.toLowerCase().indexOf(filter) > -1
                                    ? ""
                                    : "none"
                                );
                            }}
                        }}

                        function seqviz_render() {{
                            var enzymes = []

                            // check enzyme_select exists
                            var enzymeSelect = document.getElementById('enzyme_select');
                            if (enzymeSelect) {{
                                var selectedOptions = enzymeSelect.selectedOptions;
                                for (var i = 0; i < selectedOptions.length; i++) {{
                                enzymes.push(selectedOptions[i].value);
                                }}
                            }}
                            var checkboxes = document.querySelectorAll(
                                'input[name=enzyme]:checked')
                            for (var i = 0; i < checkboxes.length; i++) {{
                                enzymes.push(checkboxes[i].value)
                            }}

                            search = {{
                                query: document.getElementById('seqviz_search').value,
                                mismatch: 0 }};

                            window.seqviz
                            .Viewer("seqviz_container", {{
                                name: "{alias}",
                                seq: "{sequence}",
                                annotations: {features},
                                primers: {primers},
                                enzymes: enzymes,
                                search: search,
                                style: {{ height: "500px", width: "100%" }},
                            }})
                            .render();
                        }}
                    </script>
                    """),
                id="controls-seqviz",
            )



"""An ezcharts component for loading mosdepth data."""
import argparse
import os

import pandas as pd
from pandas.api import types as pd_types
from pkg_resources import resource_filename

from ezcharts.components.common import add_missing_windows, fasta_idx
from ezcharts.components.ezchart import EZChart
from ezcharts.components.reports.comp import ComponentReport
from ezcharts.layout.base import Snippet
from ezcharts.layout.snippets import DataTable, Tabs
from ezcharts.plots.relational import lineplot

# Categorical types
CATEGORICAL = pd_types.CategoricalDtype(ordered=True)


class DepthSummary(Snippet):
    """Generate modified bases summary plots."""

    def __init__(self, theme='epi2melabs', **kwargs):
        """Create depth summary componet.

        If dml/dmr contains results from multiple samples, each will be
        plotted in its own tab.

        :param mosdepth_summary: A path to a mosdepth summary output file or
            DataFrame
        :param mosdepth_region: A path to a mosdepth region output file or
            DataFrame
        :param faidx: A path to a fasta fai index or
            DataFrame
        """
        super().__init__(styles=None, classes=None)

        with self:
            if 'faidx' in kwargs:
                if isinstance(kwargs['faidx'], pd.DataFrame):
                    faidx = kwargs['faidx']
                elif isinstance(kwargs['faidx'], str):
                    faidx = fasta_idx(kwargs['faidx'])
                else:
                    faidx = None

            # Make modkit summary table
            if "mosdepth_summary" in kwargs:
                if isinstance(kwargs["mosdepth_summary"], pd.DataFrame):
                    summary = kwargs["mosdepth_summary"]
                else:
                    summary = load_mosdepth_summary(kwargs["mosdepth_summary"])
                # Check that the file is not empty
                if summary[0].empty:
                    raise pd.errors.EmptyDataError('Summary is empty')
                # the following assumes that `fastcat` / `bamstats` has been run
                # with `-s` which is not necessarily the case (if there was only
                # one sample) --> create a dummy 'sample_name' column if missing
                if 'sample_name' not in summary[0].columns:
                    summary[0]['sample_name'] = 'sample'
                    summary[1]['sample_name'] = 'sample'
                # If one sample, save one plot
                if len(summary[0]['sample_name'].unique()) == 1:
                    # we only got a single sample --> no dropdown
                    DataTable.from_pandas(summary[0], use_index=False)
                else:
                    # several samples --> use a dropdown menu
                    tabs = Tabs()
                    with tabs.add_dropdown_menu():
                        for sample_id, df_sample in summary[0].groupby('sample_name'):
                            with tabs.add_dropdown_tab(sample_id):
                                DataTable.from_pandas(df_sample, use_index=False)

            # Make mosdepth line plot
            if "mosdepth_region" in kwargs:
                if isinstance(kwargs["mosdepth_region"], pd.DataFrame):
                    bed = kwargs["mosdepth_region"]
                else:
                    bed = load_mosdepth_regions(
                        kwargs["mosdepth_region"], faidx=faidx,
                        winsize=25000, min_size=10000000)
                # Check that the file is not empty
                if bed.empty:
                    raise pd.errors.EmptyDataError('Region DF is empty')
                # the following assumes that `fastcat` / `bamstats` has been run
                # with `-s` which is not necessarily the case (if there was only
                # one sample) --> create a dummy 'sample_name' column if missing
                if 'sample_name' not in bed.columns:
                    bed['sample_name'] = 'sample'
                # If one sample, save one plot
                if len(bed['sample_name'].unique()) == 1:
                    # we only got a single sample --> no dropdown
                    plt = lineplot(
                        data=bed, x='total_mean_pos',
                        y='depth', hue='chrom')
                    # Remove markers
                    for s in plt.series:
                        s.showSymbol = False
                    EZChart(plt, 'epi2melabs')
                else:
                    # several samples --> use a dropdown menu
                    tabs = Tabs()
                    with tabs.add_dropdown_menu():
                        for sample_id, df_sample in bed.groupby('sample_name'):
                            with tabs.add_dropdown_tab(sample_id):
                                # we only got a single sample --> no dropdown
                                plt = lineplot(
                                    data=df_sample, x='total_mean_pos',
                                    y='depth', hue='chrom')
                                # Remove markers
                                for s in plt.series:
                                    s.showSymbol = False
                                EZChart(plt, 'epi2melabs')


# Load region mosdepth output file
def load_mosdepth_regions(
        mosdepth, faidx=None, subset=None,
        karyo=None, winsize=25000, min_size=0):
    """Load mosdepth results into dataframe.

    Input is either a single directory with all the mosdepth outputs or
    a single mosdepth file.

    Optional inputs can be:
    - faidx: A faidx dataframe with 'chrom' 'length' cols (add missing intervals)
    - winsize: A window size (int, needed to add back missing intervals)
    - karyo: A ordered karyotype (dictionary of "chrom" : int(position), e.g.
    {'chr1': 1, ..., 'chrY': 24})
    - subset: A subset of sequences to use as a list
    """
    relevant_stats_cols_dtypes = {
        "chrom": CATEGORICAL,
        "start": int,
        "end": int,
        "depth": float,
    }
    # Check inputs
    dfs = []
    if os.path.isdir(mosdepth):
        input_files = [(mosdepth, i) for i in os.listdir(mosdepth)]
    elif os.path.isfile(mosdepth):
        input_files = [(None, mosdepth)]
    else:
        raise Exception(f'No valid input: {mosdepth}')
    # If no files, raise error
    if len(input_files) == 0:
        raise FileNotFoundError(f'No valid input found in {mosdepth}')
    # Process each file
    for (inpath, fname) in input_files:
        try:
            # Define file name
            mosdepth_file = fname if not inpath else f'{inpath}/{fname}'
            # Load input
            df = pd.read_csv(
                mosdepth_file,
                sep="\t",
                names=relevant_stats_cols_dtypes.keys(),
                dtype=relevant_stats_cols_dtypes
            )
            # If it's empty, add an empty DF
            if df.empty:
                cols = relevant_stats_cols_dtypes.update(
                    {
                        'mean_pos': int, 'step': int,
                        "total_mean_pos": int, "filename": str})
                dfs.append(pd.DataFrame(columns=cols))
                continue
            # If chrom sizes are provided, add missing windows in each chrom
            if isinstance(faidx, pd.DataFrame) and isinstance(winsize, int):
                if 'length' not in faidx.columns:
                    raise ValueError(
                        'No "length" column found.',
                        'Import the fai with fasta_idx() and try again.')
                if 'chrom' not in faidx.columns:
                    raise ValueError(
                        'No "chrom" column found.',
                        'Import the fai with fasta_idx() and try again.')
                faidx = faidx.loc[faidx['length'] > min_size]
                df = add_missing_windows(df, faidx, value='depth', winsize=winsize)
            # If karyo provided, sort by given order
            df = (
                df.eval("mean_pos = (start + end) / 2")
                .eval("step = end - start")
                .reset_index(drop=True)
                )
            if subset:
                subset = subset if isinstance(subset, list) else [subset]
                df = df.loc[df['chrom'].isin(subset)]
            if karyo:
                # Sort column by chromosome number
                df = df.eval("chr_id = chrom.map(@karyo)") \
                    .sort_values(["chr_id", "start"]) \
                    .drop(columns="chr_id")
            # Remove and reorder categories
            df['chrom'] = \
                df.chrom.cat.remove_unused_categories()
            df['chrom'] = \
                df.chrom.cat.reorder_categories(
                [i for i in df.chrom.unique()])

            # If faidx is provided, use that to compute the cumulative positions
            if isinstance(faidx, pd.DataFrame):
                ref_lengths = faidx.groupby(
                    "chrom", observed=True, sort=False)["length"].last()
                total_ref_starts = ref_lengths.cumsum().shift(1, fill_value=0)
            # Otherwise, use the df
            else:
                ref_lengths = df.groupby(
                    "chrom", observed=True, sort=False)["end"].last()
                total_ref_starts = ref_lengths.cumsum().shift(1, fill_value=0)

            # Add cumulative depth
            df["total_mean_pos"] = df.apply(
                lambda x: x.mean_pos + total_ref_starts[x.chrom], axis=1)
            # Add filename
            df['filename'] = fname.split('/')[-1]
            # Add to list
            dfs.append(df)
        except pd.errors.EmptyDataError:
            cols = relevant_stats_cols_dtypes.update({'filename': str})
            dfs.append(pd.DataFrame(columns=cols)) \
                .astype({"chrom": CATEGORICAL})
    return pd.concat(dfs).reset_index(drop=True)


def load_mosdepth_summary(summary):
    """Load mosdepth results into dataframe."""
    relevant_stats_cols_dtypes = {
        "chrom": CATEGORICAL,
        "length": int,
        "bases": int,
        "mean": float,
        "min": int,
        "max": int
    }
    # Check inputs
    dfs_tot = []
    dfs_reg = []
    if os.path.isdir(summary):
        input_files = [(summary, i) for i in os.listdir(summary)]
    elif os.path.isfile(summary):
        input_files = [(None, summary)]
    else:
        raise Exception(f'No valid input: {summary}')
    # If no files, raise error
    if len(input_files) == 0:
        raise FileNotFoundError(f'No valid input found in {summary}')
    # Process each file
    for (inpath, fname) in input_files:
        try:
            # Define file name
            summary_file = fname if not inpath else f'{inpath}/{fname}'
            # Load input
            df = pd.read_csv(
                summary_file,
                sep="\t",
                usecols=relevant_stats_cols_dtypes.keys(),
                dtype=relevant_stats_cols_dtypes
                )
            # If it's empty, add an empty DF
            if df.empty:
                cols = relevant_stats_cols_dtypes.update(
                    {'filename': str})
                dfs_tot.append(pd.DataFrame(columns=cols))
                dfs_reg.append(pd.DataFrame(columns=cols))
                continue
            # Add filename
            df['filename'] = fname.split('/')[-1]
            # Split region/total stats
            dfs_tot.append(df[~df['chrom'].str.contains("_region")])
            dfs_reg.append(df[df['chrom'].str.contains("_region")])

        # Empty data error > append empty DF with the right structure
        except pd.errors.EmptyDataError:
            cols = relevant_stats_cols_dtypes.update({'filename': str})
            dfs_tot.append(pd.DataFrame(columns=cols))
            dfs_reg.append(pd.DataFrame(columns=cols))
    return (
        pd.concat(dfs_tot).reset_index(drop=True),
        pd.concat(dfs_reg).reset_index(drop=True))


def main(args):
    """Entry point to demonstrate a depth component."""
    comp_title = 'Mosdepth results'
    seq_sum = DepthSummary(
        mosdepth_summary=args.mosdepth_summary,
        mosdepth_region=args.mosdepth_region,
        faidx=args.faidx
        )
    report = ComponentReport(comp_title, seq_sum)
    report.write(args.output)


def argparser():
    """Argument parser for entrypoint."""
    parser = argparse.ArgumentParser(
        'Modkit summary',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        add_help=False)
    parser.add_argument(
        "--mosdepth_summary",
        default=resource_filename('ezcharts', "data/test/test_mosdepth_summary.tsv"),
        help="Resulting summary file from mosdepth.")
    parser.add_argument(
        "--mosdepth_region",
        default=resource_filename('ezcharts', "data/test/test_mosdepth.bed.gz"),
        help="Resulting region coverage bed file from mosdepth.")
    parser.add_argument(
        "--faidx",
        default=resource_filename('ezcharts', "data/test/ref.fa.fai"),
        help="Reference fasta fai file.")
    parser.add_argument(
        "--output",
        default="mosdepth_report.html",
        help="Output HTML file.")
    return parser



"""Test functions in utils."""

from pkg_resources import resource_filename
import pytest

from ezcharts.components import bcfstats


@pytest.mark.parametrize(
    "fname, expected_idd_empty", [
        ("data/test/bcftools/variants.stats", False),
        ("data/test/bcftools/cw-3767-variants.stats", True)
    ])
def test_001_read_bcfstats(fname, expected_idd_empty):
    """Reading a bcfstats file and checking contents."""
    fpath = resource_filename("ezcharts", fname)
    parsed_data = bcfstats.parse_bcftools_stats(fpath)

    # Check that 'IDD' key exists in parsed_data
    assert "IDD" in parsed_data

    # Check if 'IDD' is a dataframe and verify its content based on expected_idd_empty
    idd = parsed_data["IDD"]
    assert hasattr(idd, "empty")

    if expected_idd_empty:
        assert idd.empty
    else:
        assert not idd.empty



"""Run ezCharts demo."""
import argparse
import json

from bokeh.models.tickers import SingleIntervalTicker
from dominate.tags import div, h4, p
import numpy as np
import pandas as pd
from pkg_resources import resource_filename

import ezcharts as ezc
from ezcharts import util
from ezcharts.components.common import fasta_idx, HSA_CHROMOSOME_ORDER
from ezcharts.components.dss import load_dml, load_dmr
from ezcharts.components.ezchart import EZChart
from ezcharts.components.fastcat import load_bamstats_flagstat, load_stats
from ezcharts.components.fastcat import SeqCompare, SeqSummary
from ezcharts.components.modkit import load_bedmethyl, load_modkit_summary
from ezcharts.components.mosdepth import load_mosdepth_regions, load_mosdepth_summary
from ezcharts.components.nextclade import NextClade, NXTComponent
from ezcharts.components.plotmetadata import PlotMetaData
from ezcharts.components.reports.labs import LabsReport
from ezcharts.components.theme import LAB_head_resources
from ezcharts.layout.snippets import DataTable
from ezcharts.layout.snippets import Grid
from ezcharts.layout.snippets import Progress
from ezcharts.layout.snippets import Stats
from ezcharts.layout.snippets import Tabs
from ezcharts.layout.snippets.cards import Cards
from ezcharts.layout.snippets.cards import ICard
from ezcharts.layout.snippets.offcanvas import IOffCanvasClasses, OffCanvas
from ezcharts.plots import BokehPlot, Plot
from ezcharts.plots.ideogram import ideogram
from ezcharts.plots.karyomap import karyomap


# Setup simple globals
WORKFLOW_NAME = 'wf-template'
WORKFLOW_VERSION = "v1.0.0"
REPORT_TITLE = f'{WORKFLOW_NAME}-report'


def main(args):
    """Demo for ezCharts functionality."""
    logger = util.get_named_logger("ezChrtDemo")

    # Good location to prepare plots
    logger.info('Building plots')

    # Example data
    raw_df = pd.DataFrame({
        # as laid out in echarts docs for a dataset
        'product': [
            'Matcha Latte', 'Milk Tea', 'Cheese Cocoa', 'Walnut Brownie'],
        '2015': [43.3, 83.1, 86.4, 72.4],
        '2016': [85.8, 73.4, 65.2, 53.9],
        '2017': [93.7, 55.1, 82.5, 39.1]
    })
    # how we'd likely normally have it
    example_df = raw_df.melt(
        id_vars=['product'], value_vars=['2015', '2016', '2017'],
        var_name='year', value_name='sales')

    def example_plot(style="line") -> Plot:
        """Create example plot."""
        if style == "line":
            plot = ezc.lineplot(data=example_df, x='year', y='sales', hue='product')
            # Bokeh attempts to plot years in line plots as continuous variables, so
            # we get xtick labels such as 2005,2005.5. This can be fixed by setting the
            # interval on the ticker to 1.
            plot._fig.xaxis.ticker = SingleIntervalTicker(interval=1)
        elif style == "scatter":
            plot = ezc.scatterplot(data=example_df, x='year', y='sales', hue='product')
            plot._fig.xaxis.ticker = SingleIntervalTicker(interval=1)
        elif style == 'histogram':
            hist_data = pd.DataFrame(
                        [np.random.normal(loc=10.0, size=1000),
                         np.random.normal(loc=12.0, size=1000),
                         np.random.normal(loc=14.0, size=1000)]).T
            hist_data.columns = ['sample1', 'sample2', 'sample3']
            plot = ezc.histplot(hist_data, bins=50, stat='proportion')

            plot._fig.xaxis.axis_label = 'Read length'
            plot._fig.yaxis.axis_label = 'Number of reads'
        elif style == "bar":
            plot = ezc.barplot(data=example_df, x='year', y='sales', hue='product')
        elif style == "heatmap":
            # make data a matrix like you would expect for a heatmap
            df = raw_df.set_index([raw_df.columns[0]])
            # set the names of the axes
            df = df.rename_axis("Product")
            df = df.rename_axis("Year", axis=1)
            plot = ezc.heatmap(data=df)
        elif style == "boxplot":
            # lets add some data to show off outliers
            raw_df['2024'] = [10.1, 100.0, 101.1, 99.8]
            df = pd.melt(raw_df, value_vars=['2015', '2016', '2017', '2024'])
            plot = ezc.boxplot(
                df, x="variable", y="value", order=['2024', '2017', '2015', '2016'])
        else:
            raise ValueError("Unknown plot style")

        title_text = f"Example {style} chart"

        # adding a title is different for echarts and bokeh
        if isinstance(plot, BokehPlot):
            plot._fig.title.text = title_text
        else:
            plot.title = {"text": title_text}

        return plot

    # Then, time to construct a report
    logger.info('Building report')

    # Example data
    params = resource_filename('ezcharts', "data/test/params.json")
    versions = resource_filename('ezcharts', "data/test/versions.txt")
    nxt_json = resource_filename('ezcharts', "data/test/nextclade.json")
    histogram_stats_dir = tuple(
        [
            resource_filename("ezcharts", "data/test/histogram_stats/sample_2"),
            resource_filename("ezcharts", "data/test/histogram_stats/sample_1"),
            resource_filename("ezcharts", "data/test/histogram_stats/empty_sample/"),
        ]
    )

    # Create report
    # Note we need to add nextclade as a resource
    report = LabsReport(
        REPORT_TITLE, WORKFLOW_NAME, params, versions,
        workflow_version=WORKFLOW_VERSION,
        head_resources=[*LAB_head_resources, NXTComponent])

    # Add a header badge via a method on the report element.
    report.add_badge('Test badge', bg_class="bg-primary")

    # Make a progress bar for the median accuracy box
    median_accuracy = Progress(
            value_min=0,
            value_max=100,
            value_now=98.9,
            bar_cls="progress-bar-striped",
            height=50)

    # Add something directly to main_content.
    with report.main_content:
        # Stats is a snippet, but we could also just add any html
        # tags here. Snippets are little reusable portions of html,
        # styles and or scripts in python class format. They often
        # expose methods for adding content to them in a simple way.
        # E.g.
        Stats(
            columns=3,
            items=[
                ('1213986', 'Read count'),
                (median_accuracy, 'Median accuracy'),
                ('10213 bp', 'Some other stat')
            ])

    # Add sequence summary component
    with report.add_section('Sequence summaries', 'Summaries'):
        SeqSummary(
            seq_summary=histogram_stats_dir,
            color="#2a98b7",
            sample_names=tuple(['sample_2', 'sample_1', 'sample_3']))
    with report.add_section('Compare Summaries', 'Compare'):
        SeqCompare(
            histogram_stats_dir,
            sample_names=tuple(['sample_2', 'sample_1', 'sample_3']),
            alignment_stats=False,
            color="#2a98b7",
        )

    # This also adds to main_content, but provides a nice
    # container snippet as a starting context.

    with report.add_section('Controls', 'Controls'):
        # This is an example of how to use bootstrap cards
        Cards(
            columns=2,
            items=[
                ICard(
                    head="Positive",
                    body="The positive passed",
                    footer="All postive controls within threshold",
                    classes="bg-success text-white"),
                ICard(
                    head="No Template Control",
                    body="The NTC failed",
                    footer="NTC outside set thresholds.",
                    classes="bg-danger text-white")])
        OffCanvas(
            label="More details (Offcanvas demo)",
            title="Details of Controls",
            body="You can put more details here....")

    # Add sample QC cards
    with report.add_section('QC', 'QC'):
        body_content = div(
            div("Column 1 Text", _class="col"),
            div("Column 2 Text", _class="col"),
            div("Column 3 Text", _class="col"),
            _class="row",
        )
        offcanvas_classes = IOffCanvasClasses()
        offcanvas_classes.offcanvas_button = "btn btn-detail float-end"

        sample_body = div(
            div(
                h4("Sample1"),
                p("This is a sample body.")
            )
        )
        ocps = {
            'label': "View details",
            'title': 'Sample detail',
            'body': sample_body,
            'classes': offcanvas_classes
        }
        Cards(
            columns=1,
            items=[
                ICard(
                    alias="Sample1",
                    barcode="barcode01",
                    body=body_content,
                    footer="All postive controls within threshold",
                    status='pass',
                    offcanvas_params=ocps),
                ICard(
                    alias="Sample2",
                    barcode="barcode02",
                    body="Test string",
                    status='warn',
                    offcanvas_params=ocps),
                ICard(
                    alias="Sample3",
                    barcode="barcode03",
                    body=body_content,
                    footer="NTC outside set thresholds.",
                    status='fail')
                    ])

    # This also adds to main_content, but provides a nice
    # container snippet as a starting context.
    with report.add_section('Alignment results', 'Results'):
        # This is the tabbed section with ezcharts!
        tabs = Tabs()
        with tabs.add_tab('Summary'):
            # Grids are snippets that provide responsive
            # layouts via css grid
            with Grid():
                EZChart(example_plot("line"))
                EZChart(example_plot("scatter"))
                EZChart(example_plot("bar"))
                EZChart(example_plot("histogram"))
                EZChart(example_plot("boxplot"))
        with tabs.add_tab('Accuracy'):
            p("This is a mixed tab!")
            EZChart(example_plot())
        with tabs.add_tab('Depth'):
            p('Testing, testing, 1 2 3')

        # Dropdowns are nested one more level
        with tabs.add_dropdown_menu('Bar charts', change_header=False):
            with tabs.add_dropdown_tab('Simple'):
                # the simplest barplot
                plot = ezc.barplot(
                    data=example_df, x='year', y='sales'
                )
                EZChart(plot)
            with tabs.add_dropdown_tab('Simple - single colour'):
                # set color=grey for a uniform colouring
                plot = ezc.barplot(
                    data=example_df, x='year', y='sales', color='grey'
                )
                EZChart(plot)
            with tabs.add_dropdown_tab('Simple - additional styles'):
                # set margin to auto for horizontally centre plot
                plot = ezc.barplot(
                    data=example_df, x='year', y='sales'
                )
                EZChart(
                    plot, height="500px", width="500px",
                    additional_styles={"margin": "0 auto"})
            with tabs.add_dropdown_tab('Simple - single colour, ignore stacking'):
                # set color=grey for a uniform colouring
                # try setting also dodge=False, to check it doesn't break
                plot = ezc.barplot(
                    data=example_df, x='year', y='sales', color='grey', dodge=False
                )
                EZChart(plot)
            with tabs.add_dropdown_tab('Grouped'):
                # set dodge=True for a grouped barplot
                plot = ezc.barplot(
                    data=example_df, x='year', y='sales', hue='product', dodge=True
                )
                EZChart(plot)
            with tabs.add_dropdown_tab('Stacked'):
                # set dodge=False for a stacked barplot
                plot = ezc.barplot(
                    data=example_df, x='year', y='sales', hue='product', dodge=False
                )
                EZChart(plot)
            with tabs.add_dropdown_tab('Nested'):
                # set nested_x=True for a nested barplot
                plot = ezc.barplot(
                    data=example_df, x='year', y='sales', hue='product', nested_x=True
                )
                EZChart(plot)
        with tabs.add_tab('Plotting Failed'):
            # Force fail in plotting
            EZChart(ezc.sunburst([], label_rotate="tangential", label_minAngle=100))

    with report.add_section('Nextclade results', 'Nextclade', True):
        NextClade(nxt_json)

    with report.add_section('Heatmap', 'Heatmap'):
        EZChart(example_plot("heatmap"))

    with report.add_section('Human Genome', 'Genome'):
        EZChart(
            ideogram(blocks='cytobands'), height="800px")

    with report.add_section('Human karyotype heatmap', 'Karyomap'):
        vals = pd.read_csv(
            resource_filename('ezcharts', "data/test/karyomap_vals.tsv.gz"), sep='\t')
        faidx = fasta_idx(resource_filename('ezcharts', "data/test/ref.fa.fai"))
        EZChart(karyomap(
            vals, 'chr', 'pos', 'values',
            stats='count',
            ref_lengths=faidx,
            order=HSA_CHROMOSOME_ORDER))

    with report.add_section('Sunburst', 'Sunburst'):
        lineages = [{'name': 'Bacteria', 'value': 600, 'children': [
            {'name': 'Firmicutes', 'value': 600, 'children': [
                {'name': 'Bacilli', 'value': 600, 'children': [
                    {'name': 'Bacillales', 'value': 600, 'children': [
                        {'name': 'Bacillaceae', 'value': 500, 'children': [
                            {'name': 'Bacillus', 'value': 500, 'children': [
                                {'name': 'Bacillus subtilis', 'value': 250},
                                {'name': 'Bacillus aquiflavi', 'value': 250}]}]},
                        {'name': 'Staphylococcaceae', 'value': 100, 'children': [
                            {'name': 'Staphylococcus', 'value': 100, 'children': [
                                {'name': 'Staphylococcus saccharolyticus',
                                    'value': 100}]}]}]}]}]}]}]

        EZChart(ezc.sunburst(
            lineages, label_rotate="tangential", label_minAngle=100))

    with report.add_section('Table', 'Table'):
        tabs = Tabs()
        with tabs.add_tab('Table'):
            df = pd.DataFrame(np.random.random((4, 4))) * 10
            df.columns = pd.MultiIndex.from_product(
                [['sample1', 'sample2'], ['mapped', 'unmapped']])
            df['chr'] = [1, 2, 3, 4]
            df = df.set_index('chr', drop=True)
            DataTable.from_pandas(df, use_index=True)
        with tabs.add_tab('Export'):
            df_simple = pd.DataFrame(np.random.random((4, 2))) * 10
            df_simple.columns = ['sample1', 'sample2']
            DataTable.from_pandas(df_simple, export=True, file_name='samples')

    with report.add_section('Sankey plot', 'Sankey'):
        sankey_json = resource_filename('ezcharts', "data/test/sankey.json")
        with open(sankey_json, 'r') as sfh:
            sankey_dict = json.load(sfh)
            ezc.metagenomics_sankey(sankey_dict)

    with report.add_section('Plasmid', 'Plasmid'):
        plannotate_json = resource_filename(
            'ezcharts', "data/test/seqviz/plannotate.json")
        fasta = resource_filename(
            'ezcharts', "data/test/seqviz/barcode01_q10.final.fasta")
        ezc.seqviz(plannotate_json, fasta, "sample_1")

    with report.add_section('Multiple sequence alignment', 'MSA'):
        msa_file = resource_filename('ezcharts', "data/test/msa/HIGD2A.fa")
        tabs = Tabs()
        with tabs.add_tab('Full MSA'):
            ezc.msa(msa_file, color_scheme="Blossom", identity=80)
        with tabs.add_tab('Partial MSA'):
            ezc.msa(msa_file, identity=50, start=30, end=80)

    dataset_examples = {
        'fastcat': load_stats(resource_filename(
            'ezcharts',
            "data/test/fastcat/f1.tsv.gz")),
        'bamstats': load_stats(resource_filename(
            'ezcharts',
            "data/test/bamstats/bamstats.readstats.tsv.gz")),
        'flagstat': load_bamstats_flagstat(resource_filename(
            'ezcharts', "data/test/bamstats.flagstat.tsv")),
        'modkit summary': load_modkit_summary(resource_filename(
            'ezcharts', "data/test/test_modkit_summary.tsv")),
        'modkit bedMethyl': load_bedmethyl(resource_filename(
            'ezcharts', "data/test/test_modkit.bed.gz")),
        'DSS DML': load_dml(resource_filename(
            'ezcharts', "data/test/test_dml.tsv.gz")),
        'DSS DMR': load_dmr(resource_filename(
            'ezcharts', "data/test/test_dmr.tsv.gz")),
        'mosdepth summary': load_mosdepth_summary(resource_filename(
            'ezcharts', "data/test/test_mosdepth_summary.tsv")),
        'mosdepth regions': load_mosdepth_regions(resource_filename(
            'ezcharts', "data/test/test_mosdepth.bed.gz")),
        'fai index': fasta_idx(resource_filename(
            'ezcharts', "data/test/ref.fa.fai")),
    }
    with report.add_section('Data types', 'Data'):
        p(
            "The workflow comes with a range of data loaders, as shown in the",
            "tables below (10 rows max displayed).")
        tabs = Tabs()
        with tabs.add_dropdown_menu():
            for data_type in dataset_examples.keys():
                df = dataset_examples[data_type]
                with tabs.add_dropdown_tab(data_type):
                    if isinstance(df, tuple) or isinstance(df, tuple):
                        for d in df:
                            DataTable.from_pandas(d.head(10))
                    else:
                        DataTable.from_pandas(df.head(10))
    with report.add_section("Bokeh", "Bokeh"):
        plot = BokehPlot()
        plot._fig.circle(
            [1, 2, 3, 4, 5],
            [6, 7, 2, 4, 5],
            size=20,
            color="navy",
            alpha=0.5,
        )
        plot._fig.title = "Bokeh plot title"
        EZChart(plot)

    with report.add_section('Per Sample Read Counts', 'Read Counts'):
        PlotMetaData(
            resource_filename('ezcharts', "data/test/metadata.json"))

    logger.info('Reticulating splines')
    report.write(args.output)


def argparser():
    """Argument parser for entrypoint."""
    parser = argparse.ArgumentParser(
        "ezcharts demo",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        add_help=False)
    parser.add_argument(
        "--output", default="ezcharts_demo_report.html",
        help="Output HTML file.")
    return parser



"""Ideoplots."""

import argparse

import pandas as pd
from pkg_resources import resource_filename

from ezcharts.components.reports.comp import ComponentReport
from ezcharts.plots import Plot, util


class MakeRectangles(util.JSCode):
    """
    Make JSCode rectangles.

    Where:
        x left and right are given in data
        y is a categoric variable
        height is a proportion of plot-category height
    """

    def __init__(self, height: float = 0.7):
        """Instantiate the class with some jscode and some params.

        :param height: height of rectangles as proportion
            of y-axis category height.
        """
        self.height = height

        jscode = """function renderItem(params, api) {
            var categoryIndex = api.value(0);
            var start = api.coord([api.value(1), categoryIndex]);
            var end = api.coord([api.value(2), categoryIndex]);
            var height = api.size([0, 1])[1] * @height;
            var rectShape = echarts.graphic.clipRectByRect(
                {
                    x: start[0],
                    y: start[1] - height / 2,
                    width: end[0] - start[0],
                    height: height
                },
                {
                    x: params.coordSys.x,
                    y: params.coordSys.y,
                    width: params.coordSys.width,
                    height: params.coordSys.height
                }
            );
            return (
                rectShape && {
                    type: 'rect',
                    transition: ['shape'],
                    shape: rectShape,
                    style: api.style()

                }
            );}""".replace('@height', str(self.height))

        super().__init__(jscode)


def load_ucsc_bands(genome="hg38", types=None):
    """Load "standard" UCSC bands from included BED file.

    :param genome:
    :param types: a list of (stain) types to include.
    """
    # TODO: clean up package data
    src = resource_filename(
        "ezcharts",
        f"data/reference/{genome}/cytoBand.txt.gz")
    if types is None:
        types = [
            'acen', 'gneg',
            'gpos25', 'gpos50', 'gpos75', 'gpos100',
            'gvar', 'stalk']

    names = ['chr', 'start', 'end', 'name', 'type']
    bands_df = pd.read_csv(src, sep="\t", header=None, names=names)
    # annotate with colours, this is why this is isn't a generic
    # function - we assume a set of types.
    types = dict(
        acen=util.Colors.cinnabar,
        gneg=util.Colors.white,
        gpos25=util.Colors.grey90,
        gpos50=util.Colors.grey70,
        gpos75=util.Colors.grey40,
        gpos100=util.Colors.grey20,
        gvar=util.Colors.cerulean,
        stalk=util.Colors.sandstorm)

    bands_df['color'] = bands_df['type'].map(types)
    bands_df['chr'] = bands_df['chr'].str.replace('chr', '')
    bands_df = bands_df[bands_df['type'].isin(types)]
    return bands_df


def load_chr_sizes(genome="hg38"):
    """Load chr size data.

    :param genome:
    """
    # TODO: clean up package data
    src = resource_filename(
        "ezcharts",
        f"data/reference/{genome}/{genome}.chrom.sizes.gz")
    sizes = pd.read_csv(src, sep="\t", header=None, names=['chr', 'size'])
    sizes['chr'] = sizes['chr'].str.replace('chr', '')
    return sizes


def ideogram(blocks=None, track=None, genome='hg38'):
    """Draw an ideogram in various styles.

    :param blocks:
    :param track:
    """
    # TODO: a legend?

    if blocks is not None and not isinstance(blocks, (list, tuple)):
        blocks = [blocks]

    blocks_list = list()

    # find known 'blocks' datasets, actually these are currently all
    # just subsets of the bands dataset loaded as a single thing
    # and coloured by the band name.
    named_dsets = [x for x in blocks if isinstance(x, str)]
    if 'cytobands' in named_dsets:
        named_dsets = None
    if named_dsets is None or len(named_dsets):
        blocks_list.append(
            load_ucsc_bands(genome, named_dsets))

    # other blocks items given as dataframe with some required fields
    for user_dset in (x for x in blocks if isinstance(x, pd.DataFrame)):
        req_fields = pd.Series(['chr', 'start', 'end', 'color'])
        if req_fields.isin(user_dset.columns).all():
            dset = user_dset.copy()
            dset['chr'] = dset['chr'].str.replace('chr', '')
            blocks_list.append(dset)
        else:
            raise TypeError(
                "Provided blocks dataframe does not contain required fields: "
                f"{req_fields.tolist()}. Found {user_dset.columns.tolist()}")

    if track is not None:
        # note, no colour -- this is hardcoded in visual map
        # TODO: don't do this
        req_fields = pd.Series(['chr', 'start', 'value'])
        if req_fields.isin(track.columns).all():
            track_data = track.copy()
            track_data['chr'] = track_data['chr'].str.replace('chr', '')
        else:
            raise TypeError(
                "Provided trakcs dataframe does not contain required fields: "
                f"{req_fields.tolist()}")

    # need chrom sizes for outer boxes
    sizes = load_chr_sizes(genome)

    # off we go
    plt = Plot()

    # setup some variables that define our grid
    if track is None:
        top_increment, height = 4, 4
    else:
        top_increment, height = 2.0, 1.7

    grid = list()
    xaxis = list()
    yaxis = list()
    visual_map = list()
    chr_sizes = list()
    top = 1

    # this keeps count of total datasets added
    master_dataset_index = 0
    for count, (index, row) in enumerate(sizes.iterrows()):
        chromosome, chrom_size = row['chr'], row['size']
        # only consider main assembly
        if "_" in row['chr']:
            continue

        chr_sizes.append({
            'value': [chromosome, 0, chrom_size, chrom_size],
            'itemStyle': {
                'borderJoin': 'round',
                'color': 'rgba(0, 0, 0, 0)',
                'borderColor': util.Colors.not_black,
                'borderWidth': 5}})

        grid.append({
            'left': '3.5%',
            'top': f'{top}%',
            'width': '90%',
            'height': f'{height}%'})

        top += top_increment

        if track is not None:
            grid.append({
                'left': '3.5%',
                'top': f'{top}%',
                'width': '90%',
                'height': f'{height}%'})

            top += top_increment + 0.05

            # These are the grid and x/yaxis index for tracks - i.e odd
            track_grid_index = 2 * count if count != 0 else 0

            # This is the grid index and x/yaxis for chromosomes - i.e. even
            chromosome_grid_index = (2 * count) + 1 if count != 0 else 1

            # track axes
            xaxis.append({
                'gridIndex': track_grid_index,
                'show': False,
                'max': max(sizes['size'])})

            yaxis.append({
                'gridIndex': track_grid_index,
                'min': -2.2,
                'max': 2.2,
                'type': 'value',
                'position': 'right',
                'interval': 2.2,
                'axisLine': {'onZero': False}})

            # subset our track data to the chromosome we are dealing with
            data = track_data.loc[
                track_data['chr'] == chromosome]
            data = data[['start', 'value']].values.tolist()

            # track data series
            plt.add_series(dict(
                symbolSize=2,
                type='scatter',
                itemStyle={'color': util.Colors.grey60},
                xAxisIndex=track_grid_index,
                yAxisIndex=track_grid_index,
                datasetIndex=master_dataset_index))
            plt.add_dataset(dict(source=data))

            # TODO: ONE FOR LATER, this is not generic
            visual_map.append(dict(
                left='right',
                show=False,
                min=-1,
                max=1,
                seriesIndex=master_dataset_index,
                inRange={
                    'color': [
                        util.Colors.cerulean,
                        util.Colors.grey60,
                        util.Colors.cinnabar]},
                text=['>0', '<0'],
                calculable=False))

            # increment our dataset index
            master_dataset_index += 1
        else:  # adding track
            chromosome_grid_index = count

        # now add blocks
        for blocks_data in blocks_list:
            # subset our bands data to the chormosome we are dealing with
            blocks_data_chr = blocks_data.loc[
                blocks_data['chr'] == chromosome]

            blocks_dataset = dict()
            for index_blocks, row_blocks in blocks_data_chr.iterrows():
                if row_blocks['color'] not in blocks_dataset:
                    blocks_dataset[row_blocks['color']] = list()
                blocks_dataset[row_blocks['color']].append([
                    row_blocks['chr'],
                    row_blocks['start'],
                    row_blocks['end'],
                    row_blocks['end']-row_blocks['start']])

            for count, (data_type, data) in enumerate(blocks_dataset.items()):
                if data_type == 0:
                    continue

                plt.add_series(dict(
                    xAxisIndex=len(xaxis),
                    yAxisIndex=len(yaxis),
                    datasetIndex=master_dataset_index,
                    type='custom',
                    itemStyle={'color': data_type},
                    encode={
                        'x': [1, 2],
                        'y': 0},
                    renderItem=MakeRectangles(height=0.7)))
                plt.add_dataset(dict(source=data))

                # increment our dataset index
                master_dataset_index += 1

        # showing something?
        show = False
        if count + 1 == len(sizes):
            show = True

        # axes for chromsome ideogram
        xaxis.append({
            'gridIndex': chromosome_grid_index,
            'show': show,
            'max':  max(sizes['size'])})
        yaxis.append({
            'gridIndex': chromosome_grid_index,
            'axisLine': {'show': False},
            'axisTick': {'show': False},
            'type': 'category',
            'data': [{'value': row['chr']}]})

        # chromosome outlines
        plt.add_series(dict(
            type='custom',
            xAxisIndex=chromosome_grid_index,
            yAxisIndex=chromosome_grid_index,
            itemStyle={
                'color': 'rgba(0, 0, 0, 0)',
                'borderColor': util.Colors.not_black,
                'borderWidth': 3,
                'borderCap': 'round'},
            datasetIndex=master_dataset_index,
            encode={
                'x': [1, 2],
                'y': 0},
            renderItem=MakeRectangles(height=0.7)))
        plt.add_dataset(
            dict(source=[[row['chr'], 0, row['size'], row['size']]]))

        master_dataset_index += 1

    plt.grid = grid
    plt.xAxis = xaxis
    plt.yAxis = yaxis
    plt.visualMap = visual_map
    return plt


def main(args):
    """Plot some chromosomes."""
    blocks = list()
    if args.show_bands:
        blocks.append('cytobands')
    if args.blocks is not None:
        blk = pd.read_csv(
            args.blocks, sep="\t", header=None,
            names=['chr', 'start', 'end'])
        blk['color'] = util.Colors.grey70
        blocks.append(blk)
    track_data = None
    if args.track is not None:
        track_data = pd.read_csv(
            args.track, sep="\t", header=None,
            names=['chr', 'start', 'end', 'value'])
        track_data['value'] = track_data['value'].clip(lower=args.track_floor)

    plot = ideogram(blocks=blocks, track=track_data, genome=args.genome)
    ComponentReport.from_plot(plot, args.output, height="1000px")


def argparser():
    """Argument parser for entrypoint."""
    parser = argparse.ArgumentParser(
        'ezChart ideogram Demo',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        add_help=False)
    parser.add_argument(
        "--genome",
        default='hg38',
        choices=["hg38", "hg19"],
        help="Genome build to use")
    parser.add_argument("--show_bands", action='store_true')
    parser.add_argument(
        "--blocks",
        default=None,
        help="File to plot as rectangles on chromosomes")
    parser.add_argument(
        "--track",
        default=None,
        help="File to plot as a scatter plot")
    parser.add_argument(
        "--output",
        default="ezchart_ideogram.html",
        help="Output HTML file.")
    return parser



"""An ezcharts component make plots from the metadata.json."""

import argparse
import json

from bokeh.models import HoverTool
import pandas as pd

import ezcharts as ezc
from ezcharts.components import ezchart
from ezcharts.components.reports.comp import ComponentReport
from ezcharts.layout.base import Snippet
from ezcharts.plots.util import Colors


class PlotMetaData(Snippet):
    """A meta data plotting report component for outputs in meta_data.json."""

    TAG = 'meta-plots'

    def __init__(
            self,
            metadata_json,
            **kwargs
    ) -> None:
        """Initialize a meta plots instance."""
        super().__init__(
            styles=None,
            classes=None)

        # Read count bar plot from the counts in the JSON.
        with self:
            read_count_barplot(metadata_json, **kwargs)


def read_count_barplot(
        metadata_json,
        orient='h',
        fill_color=Colors.cerulean,
        line_color=Colors.cerulean,
        theme='epi2melabs'):
    """Plot per sample read count bar chart."""
    # Per barcode read count
    with open(metadata_json) as data_file:
        data = json.load(data_file)
    barcode_counts = pd.DataFrame(data) \
        .rename(columns={'n_seqs': 'Number of reads', 'alias': 'Sample'})
    order = barcode_counts["Sample"].tolist()
    order.sort()
    if orient == 'h':
        order = order[::-1]
        plt = ezc.barplot(
            data=barcode_counts, x='Number of reads', y='Sample',
            order=order, orient=orient,
            fill_color=fill_color, line_color=line_color)
        hover = plt._fig.select(dict(type=HoverTool))
        hover.tooltips = [("Number of reads", "@right")]
    else:
        plt = ezc.barplot(
            data=barcode_counts, x='Sample', y='Number of reads',
            order=order, orient=orient,
            fill_color=fill_color, line_color=line_color)
        hover = plt._fig.select(dict(type=HoverTool))
        hover.tooltips = [("Number of reads", "@top")]
    ezchart.EZChart(
        plt,
        theme=theme)


def main(args):
    """Entry point to create a report from meta data JSON."""
    meta_data_plots = PlotMetaData(args.metadata)
    read_count_barplot(args.metadata)
    report = ComponentReport(
        'Meta Plots', meta_data_plots)
    report.write(args.output)


def argparser():
    """Argument parser for entrypoint."""
    parser = argparse.ArgumentParser(
        'Meta Plots',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        add_help=False
    )
    parser.add_argument(
        "--metadata",
        help="Meta data json."
    )
    parser.add_argument(
        "--output",
        default="readscounts.html",
        help="Output HTML file."
    )
    return parser



"""A reusable software versions table."""
import os

from ezcharts.layout.snippets.table import DataTable


#
# These functions below were lifted from aplanat and probably need reworking,
# should probably just inherit from DataTable above.
#

class VersionsTable(DataTable):
    """Display workflow params."""

    def __init__(
        self,
        versions_path: str,
        **kwargs
    ) -> None:
        """Create table."""
        if os.path.isdir(versions_path):
            version_files = [
                os.path.join(versions_path, x)
                for x in os.listdir(versions_path)]
        elif os.path.isfile(versions_path):
            version_files = [versions_path]
        else:
            raise IOError('`versions` should be a file or directory.')
        super().__init__(['Name', 'Version'], **kwargs)
        for fname in version_files:
            try:
                with open(fname, 'r', encoding='utf-8') as fh:
                    with self.body:
                        for line in fh.readlines():
                            name, version = line.strip().split(',')
                            self.add_row(
                                title=None, columns=[name, version])
            except FileNotFoundError:
                pass


class VersionsList():
    """Display workflow params as a more compact list."""

    def __init__(
        self,
        versions_path: str,
        **kwargs
    ) -> None:
        """Create string list."""
        if os.path.isdir(versions_path):
            version_files = [
                os.path.join(versions_path, x)
                for x in os.listdir(versions_path)]
        elif os.path.isfile(versions_path):
            version_files = [versions_path]
        else:
            raise IOError('`versions` should be a file or directory.')
        result = []
        for fname in version_files:
            try:
                with open(fname, 'r', encoding='utf-8') as fh:
                    for line in fh.readlines():
                        name, version = line.strip().split(',')
                        result.append(f"{name} ({version})")
            except FileNotFoundError:
                pass
        self.ver_string = ",".join(result)

    def __repl__(self):
        """Return version list."""
        return self.ver_string



"""Get default cards layouts."""
from typing import List, NamedTuple, Optional, Type

from dominate.tags import (
    div, h5, html_tag, p, span, strong, table, td, tr)

from ezcharts.layout.base import IClasses, Snippet
from ezcharts.layout.snippets import Grid
from ezcharts.layout.snippets import OffCanvas
from ezcharts.layout.util import cls


class ICard(NamedTuple):
    """Card definition class."""

    body: object
    head: Optional[str] = None
    alias: Optional[str] = None
    barcode: Optional[str] = None
    status: Optional[str] = None
    head_msg: Optional[str] = None
    footer: Optional[str] = None
    classes: Optional[str] = None
    offcanvas_params: Optional[dict] = None


class ICardClasses(IClasses):
    """Cards html classes."""

    container: str = cls("container", "px-0")
    card: str = cls("card", "px-0", "mb-4", "shadow-sm")
    card_body: str = cls("card-body")
    card_heading: str = cls("card-header")
    card_text: str = cls("card-text")
    card_footer: str = cls("card-footer small")


class Cards(Snippet):
    """A styled section tag."""

    TAG = 'div'

    def __init__(
        self,
        columns: Optional[int] = 2,
        items: List[ICard] = [],
        classes: ICardClasses = ICardClasses(),
    ) -> None:
        """Create the grid layout."""
        super().__init__(
            styles=None,
            classes=classes,
            className=classes.container)

        with self:
            self.items = Grid(columns)

        with self.items:
            for item in items:
                self.add_card(
                    item.body,
                    item.head,
                    item.alias,
                    item.barcode,
                    item.status,
                    item.footer,
                    item.classes,
                    item.offcanvas_params
                    )

    def add_card(
        self,
        body: object,
        head: Optional[str] = None,
        alias: Optional[str] = None,
        barcode: Optional[str] = None,
        status: Optional[str] = None,
        footer: Optional[str] = None,
        card_cls: Optional[str] = None,
        offcanvas_params: Optional[dict] = None,
        head_tag: Type[html_tag] = h5,
        body_tag: Type[html_tag] = p

    ) -> None:
        """Add cards to the grid."""
        with self.items:
            with div(
                    className=cls(
                        self.classes.card,
                        card_cls if card_cls is not None else '')):

                # sample badge
                if status:
                    sample_badge_cls = [
                        'badge', 'badge-icon-solid', 'rounded-pill', 'p-2']
                    if status == 'pass':
                        sample_badge_cls.append('badge-pass-solid')
                        msg = "Pass"
                    elif status == 'warn':
                        sample_badge_cls.append('badge-unknown-solid')
                        msg = "Warning"
                    elif status == 'fail':
                        sample_badge_cls.append('badge-fail-solid')
                        msg = "Fail"
                    else:
                        msg = status
                    _span = span(msg, cls=' '.join(sample_badge_cls))

                # rest of heading
                with div(className=self.classes.card_heading):
                    if head:
                        head_tag(head)
                    else:
                        with div(className="row"):
                            with div(className="col-md-3"):
                                with table():
                                    with tr(cls="col-md-3"):
                                        if alias:
                                            td(strong(alias), cls="pr-4")
                                        if barcode:
                                            td(barcode, cls="px-4")
                                        if status:
                                            td(_span)
                            if offcanvas_params is not None:
                                with div(className="col-md-9"):
                                    OffCanvas(**offcanvas_params).render()

                # add body
                with div(className=self.classes.card_body):
                    try:
                        body_tag(body, className=self.classes.card_text)
                    except TypeError:
                        body_tag(body)

                # add footer if needed
                if footer is not None:
                    div(footer, className=self.classes.card_footer)



"""Test functions in utils."""

from ezcharts import demo


def test_run_demo():
    """Run the demo, just as a test that plots all run."""
    parser = demo.argparser()
    args = parser.parse_args(["--output", "ezcharts_demo_report.html"])
    demo.main(args)



"""Test functions in utils."""

import argparse
import logging

import pandas as pd

from ezcharts import util
from ezcharts.plots.util import concat_dfs_with_categorical_columns


def test_create_logger():
    """Make a logger."""
    logger = util.get_named_logger("MyLogger")
    assert isinstance(logger, logging.Logger)


def test_set_log_level():
    """Make a parser."""
    parser = util._log_level()
    assert isinstance(parser, argparse.ArgumentParser)


def test_010_pandas_cat_properties():
    """Ensure concatenating categoricals works as intended."""
    # Input columns for fastcat
    example_cols_dtypes = {
        "string": str,
        "int": int,
        "bi_category": "category",  # all the dfs contains these categories: F,M
        "open_category": "category",  # categories aren't in all the dfs: filename
        "float": float,
        # "datetime": pd.Timestamp('20240310')
    }
    d1 = {
        "string": ['rA', 'rB', 'rC'],
        "int": [1000, 2000, 1500],
        "bi_category": ['B', 'A', 'B'],
        "open_category": ['C', 'A', 'C'],
        "float": [3.1, 9.9, 4.9]
    }
    df1 = pd.DataFrame(d1).astype(example_cols_dtypes)
    d2 = {
        "string": ['rA_2', 'rB_2', 'rC_2'],
        "int": [2000, 1000, 5000],
        "bi_category": ['A', 'A', 'A'],
        "open_category": ['B', 'A', 'B'],  # 'B' is not present in df1
        "float": [4.9, 9.9, 3.1]
    }
    df2 = pd.DataFrame(d2).astype(example_cols_dtypes)
    df_empty = pd.DataFrame(
        columns=example_cols_dtypes.keys()
        ).astype(example_cols_dtypes)

    actual = concat_dfs_with_categorical_columns(dfs=[df1, df2, df_empty])
    expected_d = {
        k1: d1[k1] + d2[k2]
        for (k1, k2) in zip(d1.keys(), d2.keys())}
    expected = pd.DataFrame(expected_d).astype(example_cols_dtypes)
    pd.testing.assert_frame_equal(
        expected, actual,
        check_dtype=True, check_categorical=True, check_exact=True
    )



"""An ezcharts component for loading modkit data."""
import argparse
import os

import pandas as pd
from pandas.api import types as pd_types
from pkg_resources import resource_filename

from ezcharts.components.common import fasta_idx, MOD_CONVERT
from ezcharts.components.ezchart import EZChart
from ezcharts.components.reports.comp import ComponentReport
from ezcharts.layout.base import Snippet
from ezcharts.layout.snippets import DataTable, Tabs
from ezcharts.plots.karyomap import karyomap

# Categorical types
CATEGORICAL = pd_types.CategoricalDtype(ordered=True)


class MKSummary(Snippet):
    """Generate modified bases summary plots."""

    def __init__(self, theme='epi2melabs', **kwargs):
        """Create depth summary componet.

        If dml/dmr contains results from multiple samples, each will be
        plotted in its own tab.

        :param modkit_summary: A path to a modkit summary output file or
            DataFrame
        :param bedmethyl: A path to a modkit bedmethyl output file or
            DataFrame
        :param faidx: A path to a fasta fai index or
            DataFrame
        """
        super().__init__(styles=None, classes=None)

        with self:
            if 'faidx' in kwargs:
                if isinstance(kwargs['faidx'], pd.DataFrame):
                    faidx = kwargs['faidx']
                elif isinstance(kwargs['faidx'], str):
                    faidx = fasta_idx(kwargs['faidx'])
                else:
                    faidx = None
            # Make modkit summary table
            if "modkit_summary" in kwargs:
                if isinstance(kwargs["modkit_summary"], pd.DataFrame):
                    summary = kwargs["modkit_summary"]
                else:
                    summary = load_modkit_summary(kwargs["modkit_summary"])
                # Check that the file is not empty
                if summary.empty:
                    raise pd.errors.EmptyDataError('Summary is empty')
                # the following assumes that `fastcat` / `bamstats` has been run
                # with `-s` which is not necessarily the case (if there was only
                # one sample) --> create a dummy 'sample_name' column if missing
                if 'sample_name' not in summary.columns:
                    summary['sample_name'] = 'sample'
                # If one sample, save one plot
                if len(summary['sample_name'].unique()) == 1:
                    # we only got a single sample --> no dropdown
                    DataTable.from_pandas(summary, use_index=False)
                else:
                    # several samples --> use a dropdown menu
                    tabs = Tabs()
                    with tabs.add_dropdown_menu():
                        for sample_id, df_sample in summary.groupby('sample_name'):
                            with tabs.add_dropdown_tab(sample_id):
                                DataTable.from_pandas(df_sample, use_index=False)
            # Make modkit bedmethyl karyomap
            if "bedmethyl" in kwargs:
                if isinstance(kwargs["bedmethyl"], pd.DataFrame):
                    bedmethyl = kwargs["bedmethyl"]
                else:
                    bedmethyl = load_bedmethyl(
                        kwargs["bedmethyl"], faidx=faidx, split_all=True)
                # Check that the file is not empty
                if bedmethyl.empty:
                    raise pd.errors.EmptyDataError('Bedmethyl is empty')
                # the following assumes that `fastcat` / `bamstats` has been run
                # with `-s` which is not necessarily the case (if there was only
                # one sample) --> create a dummy 'sample_name' column if missing
                if 'sample_name' not in bedmethyl.columns:
                    bedmethyl['sample_name'] = 'sample'
                # If one sample, save one plot
                if len(bedmethyl['sample_name'].unique()) == 1:
                    # we only got a single sample --> no dropdown
                    EZChart(karyomap(
                        bedmethyl, 'chrom', 'start',
                        'score', ref_lengths=faidx), 'epi2melabs')
                else:
                    # several samples --> use a dropdown menu
                    tabs = Tabs()
                    with tabs.add_dropdown_menu():
                        for sample_id, df_sample in bedmethyl.groupby('sample_name'):
                            with tabs.add_dropdown_tab(sample_id):
                                EZChart(karyomap(
                                    df_sample, 'chrom',
                                    'pos', 'fdr', ref_lengths=faidx), 'epi2melabs')


# Load mod bedMethyl file
def load_modkit_summary(summary_dir):
    """Load bedmethyl file.

    Input is either a single directory with all the modkit summaries or
    a single modkit summary file.
    """
    # Col names
    relevant_stats_cols_dtypes = {
        "sample": CATEGORICAL,
        "type": CATEGORICAL,
        "base": CATEGORICAL,
        "code": CATEGORICAL,
        "pass_count": int,
        "pass_frac": float,
        "all_counts": int,
        "all_frac": float,
    }
    # Prepare input files
    dfs = []
    if os.path.isdir(summary_dir):
        input_files = [(summary_dir, i) for i in os.listdir(summary_dir)]
    elif os.path.isfile(summary_dir):
        input_files = [(None, summary_dir)]
    else:
        raise Exception(f'No valid input: {summary_dir}')
    # If no files, raise error
    if len(input_files) == 0:
        raise FileNotFoundError(f'No valid input found in {summary_dir}')
    # Process each file
    for (inpath, fname) in input_files:
        try:
            f_path = fname if not inpath else f'{inpath}/{fname}'
            try:
                df = pd.read_csv(
                    f_path,
                    dtype=relevant_stats_cols_dtypes,
                    sep='\t',
                    comment='#')
            except TypeError:
                df = pd.read_csv(
                    f_path,
                    names=relevant_stats_cols_dtypes,
                    dtype=relevant_stats_cols_dtypes,
                    sep='\t',
                    comment='#')
            # If it's empty, add an empty DF
            if df.empty:
                cols = relevant_stats_cols_dtypes.update(
                    {'mod': str, 'threshold': str, 'filename': int})
                dfs.append(pd.DataFrame(columns=cols))
                continue
            # Convert modification to the appropriate code
            df['mod'] = df.apply(
                lambda x: MOD_CONVERT.get(x.code, x.base), axis=1)
            df.astype({'mod': CATEGORICAL})
            # Read and add the thresholds value
            thresholds = {}
            for line in open(f_path):
                line = line.strip().split()
                # Can be different thresholds
                if 'pass_threshold_' in line[1]:
                    thresholds[line[1].split('_')[-1]] = float(line[2])
                # no need to iterate over the whole file
                elif not line[0].startswith('#'):
                    break
            # Add corresponding threshold depending on the base
            df['threshold'] = df.apply(
                lambda x: thresholds[x.base], axis=1)
            # Add file name
            df['filename'] = fname.split('/')[-1]
            # Ensure that the mod-base code is consistent
            dfs.append(df)
        except pd.errors.EmptyDataError:
            cols = relevant_stats_cols_dtypes.update({'filename': str})
            dfs.append(pd.DataFrame(columns=cols))
    return pd.concat(dfs).reset_index(drop=True)


def load_bedmethyl(bedmethyl_input, faidx=None, split_all=False):
    """Load modkit bedmethyl file.

    Input is either a single directory with all the modkit bedmethyls or
    a single modkit bedmethyl file.

    Optional inputs can be:
    - faidx: A faidx dataframe with 'chrom' 'length' cols (add missing intervals)
    """
    # Col names
    long_cols_dtypes = {
        "chrom": CATEGORICAL,
        "start": int,
        "end": int,
        "mod": CATEGORICAL,
        "score": int,
        "strand": CATEGORICAL,
        "startp": int,
        "endp": int,
        "colour": CATEGORICAL,
        "Nvalid": int,
        "fraction": float,
        "Nmod": int,
        "Ncanonical": int,
        "Nother_mod": int,
        "Ndelete": int,
        "Nfail": int,
        "Ndiff": int,
        "Nnocall": int,
    }
    short_cols_dtypes = {
        'chrom': CATEGORICAL,
        'start': int,
        'end': int,
        'mod': str,
        'score': int,
        'strand': str,
        'start_p': int,
        'end_p': int,
        "colour": CATEGORICAL,
        'vals': str,
    }
    # Check inputs
    dfs = []
    if os.path.isdir(bedmethyl_input):
        input_files = [(bedmethyl_input, i) for i in os.listdir(bedmethyl_input)]
    elif os.path.isfile(bedmethyl_input):
        input_files = [(None, bedmethyl_input)]
    else:
        raise Exception(f'No valid input: {bedmethyl_input}')
    # If no input files are found, raise error
    if len(input_files) == 0:
        raise FileNotFoundError(f'No valid input found in {bedmethyl_input}')
    # Process each file
    for (inpath, fname) in input_files:
        try:
            bedmethyl = fname if not inpath else f'{inpath}/{fname}'
            # Load input bedmethyl
            if split_all is True:
                cols = long_cols_dtypes
                df = pd.read_csv(
                    bedmethyl,
                    sep="\t| ",
                    engine='python',
                    names=cols.keys(),
                    dtype=cols)
            else:
                cols = short_cols_dtypes
                df = pd.read_csv(
                    bedmethyl,
                    sep="\t",
                    engine='python',
                    names=short_cols_dtypes.keys(),
                    dtype=short_cols_dtypes)
            # If it's empty, add an empty DF
            if df.empty:
                cols = cols.update(
                    {'total_mean_pos': str, 'filename': int})
                dfs.append(pd.DataFrame(columns=cols))
                continue
            # If the dataframe is empty, append it directly.
            if isinstance(faidx, pd.DataFrame):
                if 'length' not in faidx.columns:
                    raise ValueError(
                        'No "length" column found.',
                        'Import the fai with fasta_idx() and try again.')
                if 'chrom' not in faidx.columns:
                    raise ValueError(
                        'No "chrom" column found.',
                        'Import the fai with fasta_idx() and try again.')
                ref_lengths = faidx.groupby(
                    "chrom", observed=True, sort=False)["length"].last()
                total_ref_starts = ref_lengths.cumsum().shift(1, fill_value=0)
            else:
                # Compute reference lenghts from the input file
                ref_lengths = df.groupby(
                    "chrom", observed=True, sort=False)["end"].last()
                # Compute cumulative position
                total_ref_starts = ref_lengths \
                    .cumsum() \
                    .shift(1, fill_value=0)
            # Compute the cumulative position.
            df["total_mean_pos"] = df.apply(
                lambda x: x.start + total_ref_starts[x.chrom], axis=1)
            # Add file name
            df['filename'] = fname.split('/')[-1]
            # Add to output list
            dfs.append(df)
        except pd.errors.EmptyDataError:
            if split_all is True:
                cols = long_cols_dtypes.update({'filename': str})
            else:
                cols = short_cols_dtypes.update({'filename': str})
            dfs.append(pd.DataFrame(columns=cols))
    return pd.concat(dfs).reset_index(drop=True)


def main(args):
    """Entry point to demonstrate a modkit component."""
    comp_title = 'ModKit results'
    seq_sum = MKSummary(
        modkit_summary=args.modkit_summary,
        bedmethyl=args.bedmethyl,
        faidx=args.faidx
        )
    report = ComponentReport(comp_title, seq_sum)
    report.write(args.output)


def argparser():
    """Argument parser for entrypoint."""
    parser = argparse.ArgumentParser(
        'Modkit summary',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        add_help=False)
    parser.add_argument(
        "--modkit_summary",
        default=resource_filename('ezcharts', "data/test/test_modkit_summary.tsv"),
        help="Resulting summary file from modkit summary.")
    parser.add_argument(
        "--bedmethyl",
        default=resource_filename('ezcharts', "data/test/test_modkit.bed.gz"),
        help="Resulting bedmethyl file from modkit pileup.")
    parser.add_argument(
        "--faidx", required=False,
        default=resource_filename('ezcharts', "data/test/ref.fa.fai"),
        help="Reference fasta fai index.")
    parser.add_argument(
        "--output",
        default="modkit_report.html",
        help="Output HTML file.")
    return parser



"""Relational plots."""

import numbers

from bokeh.models import HoverTool
from seaborn.relational import _LinePlotter, _ScatterPlotter

from ezcharts.plots import BokehPlot, util


__all__ = ["relplot", "scatterplot", "lineplot"]

DEFAULT_PALETTE = util.choose_palette()
DEFAULT_COLOR = DEFAULT_PALETTE[0]


def relplot(
        data=None, *,
        x=None, y=None, hue=None, size=None, style=None, units=None,
        row=None, col=None, col_wrap=None, row_order=None, col_order=None,
        palette=None, hue_order=None, hue_norm=None,
        sizes=None, size_order=None, size_norm=None,
        markers=None, dashes=None, style_order=None,
        legend="auto", kind="scatter", height=5, aspect=1, facet_kws=None,
        **kwargs):
    """Figure-level interface for drawing relational plots onto a FacetGrid."""
    raise NotImplementedError


class Mixin:
    """Mixin class to define how to plot lines and points."""

    def plot(self, plt: BokehPlot, kws):
        """Plot the plot."""
        # Get the potentially normalised data from seaborn
        data = self.plot_data.dropna()
        if data.empty:
            return

        if 'hue' not in data:
            # add a `hue` column (this can't be set to `None` as it would produce an
            # empty plot)
            data['hue'] = "values"

        x_name = self.variables.get("x", None)
        y_name = self.variables.get("y", None)

        symbol_size = kws.get('s', 8)
        if not isinstance(symbol_size, numbers.Number):
            raise NotImplementedError(
                "symbol size (s) currently only accepts a constant value")

        line_width = kws.get('linewidth', 3)
        if not isinstance(line_width, numbers.Number):
            raise NotImplementedError(
                "linewidth currently only accepts a constant value")

        # Available markers/glyphs can be found here:
        # https://docs.bokeh.org/en/latest/docs/reference/models/glyphs.html
        marker = kws.get('marker')
        # TODO: size and style are also grouping "semantic" variables

        y_min = 0
        for hue_name, df in data.groupby('hue'):
            relational_kwargs = {}
            if data['hue'].nunique() > 1:
                relational_kwargs["legend_label"] = hue_name

            y_min = min(y_min, df.y.min())
            color = kws.get("color")
            if self._hue_map.levels:
                # the hue map is not empty (i.e. it was initialised with a palette); use
                # this to override the color (in case one was passed)
                color = self._hue_map(hue_name)

            if self.series_type == 'line':
                plt._fig.line(
                    df.x, df.y, line_color=color,
                    line_width=line_width, **relational_kwargs)

            elif self.series_type == 'scatter':
                plt._fig.scatter(
                    df.x, df.y, color=color, **relational_kwargs)

            if marker is not False:
                if marker is None or marker is True:
                    marker = 'circle'
                # Use scatter to add the markers
                plt._fig.scatter(
                    df.x, df.y, size=symbol_size, marker=marker,
                    color=color, **relational_kwargs)

            plt._fig.xaxis.axis_label = x_name
            plt._fig.yaxis.axis_label = y_name

        # Default the y_range start to zero unless y contains negative values.
        plt._fig.y_range.start = y_min

        # TODO: add legend
        return plt


class ScatterPlotter(Mixin, _ScatterPlotter):
    """Making scatter plots."""

    series_type = 'scatter'


def scatterplot(
        data=None, *,
        x=None, y=None, hue=None, size=None, style=None,
        palette=None, hue_order=None, hue_norm=None,
        sizes=None, size_order=None, size_norm=None,
        markers=True, style_order=None, legend="auto", ax=None,
        **kwargs):
    """Draw a scatter plot with possibility of several semantic groupings."""
    # see https://github.com/mwaskom/seaborn/blob/949dec3666ab12a366d2fc05ef18d6e90625b5fa/seaborn/relational.py#L726  # noqa

    if palette is None:
        palette = DEFAULT_PALETTE

    if not kwargs.get("color"):
        kwargs["color"] = DEFAULT_COLOR

    # TODO: what of this do we actually care for
    variables = ScatterPlotter.get_semantics(locals())
    p = ScatterPlotter(data=data, variables=variables, legend=legend)
    p.map_hue(palette=palette, order=hue_order, norm=hue_norm)
    p.map_size(sizes=sizes, order=size_order, norm=size_norm)
    p.map_style(markers=markers, order=style_order)

    plt = BokehPlot(title=kwargs.get('title', ""))
    hover = plt._fig.select(dict(type=HoverTool))
    hover.tooltips = [(x, "@x"), (y, "@y")]
    p.plot(plt, kwargs)

    return plt


class LinePlotter(Mixin, _LinePlotter):
    """Making line plots."""

    series_type = 'line'


def lineplot(
        data=None, *,
        x=None, y=None, hue=None, size=None, style=None, units=None,
        palette=None, hue_order=None, hue_norm=None,
        sizes=None, size_order=None, size_norm=None,
        dashes=True, markers=None, style_order=None,
        estimator="mean", errorbar=("ci", 95), n_boot=1000, seed=None,
        orient="x", sort=True, err_style="band", err_kws=None,
        legend="auto", ci="deprecated", ax=None, **kwargs):
    """Draw a line plot with possibility of several semantic groupings."""
    # see https://github.com/mwaskom/seaborn/blob/949dec3666ab12a366d2fc05ef18d6e90625b5fa/seaborn/relational.py#L597  # noqa

    if palette is None:
        palette = DEFAULT_PALETTE

    if not kwargs.get("color"):
        kwargs["color"] = DEFAULT_COLOR

    # TODO: what of this do we actually care for
    variables = _LinePlotter.get_semantics(locals())
    p = LinePlotter(
        data=data, variables=variables, estimator=estimator, n_boot=n_boot,
        seed=seed, errorbar=errorbar, sort=sort, orient=orient,
        err_style=err_style, err_kws=err_kws, legend=legend)
    p.map_hue(palette=palette, order=hue_order, norm=hue_norm)
    p.map_size(sizes=sizes, order=size_order, norm=size_norm)
    p.map_style(markers=markers, dashes=dashes, order=style_order)

    plt = BokehPlot(title=kwargs.get('title', ""))
    hover = plt._fig.select(dict(type=HoverTool))
    hover.tooltips = [(x, "@x"), (y, "@y")]

    p.plot(plt, kwargs)
    return plt



"""An ezcharts component for loading data from DSS R package.

[DSS](https://bioconductor.org/packages/release/bioc/vignettes/DSS/inst/doc/DSS.html)
"""
import argparse
import os

import pandas as pd
from pandas.api import types as pd_types
from pkg_resources import resource_filename

from ezcharts.components.common import fasta_idx
from ezcharts.components.ezchart import EZChart
from ezcharts.components.reports.comp import ComponentReport
from ezcharts.layout.base import Snippet
from ezcharts.layout.snippets import Tabs
from ezcharts.plots.karyomap import karyomap

# Categorical types
CATEGORICAL = pd_types.CategoricalDtype(ordered=True)


class DMSummary(Snippet):
    """Generate differentially modified plots."""

    def __init__(self, theme='epi2melabs', **kwargs):
        """Create depth summary componet.

        If dml/dmr contains results from multiple samples, each will be
        plotted in its own tab.

        :param dml: A path to a DSS differentially modified loci file or
            DataFrame
        :param dmr: A path to a DSS differentially modified regions file or
            DataFrame
        :param faidx: A path to a fasta fai index or
            DataFrame
        """
        super().__init__(styles=None, classes=None)

        with self:
            if 'faidx' in kwargs:
                if isinstance(kwargs['faidx'], pd.DataFrame):
                    faidx = kwargs['faidx']
                elif isinstance(kwargs['faidx'], str):
                    faidx = fasta_idx(kwargs['faidx'])
                else:
                    faidx = None
            if "dml" in kwargs:
                if isinstance(kwargs["dml"], pd.DataFrame):
                    dml = kwargs["dml"]
                else:
                    dml = load_dml(kwargs["dml"], faidx=None)
                # Check that the file is not empty
                if dml.empty:
                    raise pd.errors.EmptyDataError('DML is empty')
                # the following assumes that `fastcat` / `bamstats` has been run
                # with `-s` which is not necessarily the case (if there was only
                # one sample) --> create a dummy 'sample_name' column if missing
                if 'sample_name' not in dml.columns:
                    dml['sample_name'] = 'sample'
                # If one sample, save one plot
                if len(dml['sample_name'].unique()) == 1:
                    # we only got a single sample --> no dropdown
                    EZChart(karyomap(
                        dml, 'chrom', 'pos', 'fdr', ref_lengths=faidx), 'epi2melabs')
                else:
                    # several samples --> use a dropdown menu
                    tabs = Tabs()
                    with tabs.add_dropdown_menu():
                        for sample_id, df_sample in dml.groupby('sample_name'):
                            with tabs.add_dropdown_tab(sample_id):
                                EZChart(karyomap(
                                    df_sample, 'chrom', 'pos',
                                    'fdr', ref_lengths=faidx), 'epi2melabs')
            if "dmr" in kwargs:
                if isinstance(kwargs["dmr"], pd.DataFrame):
                    dmr = kwargs["dmr"]
                else:
                    dmr = load_dmr(kwargs["dmr"], faidx=faidx)
                # Check that the file is not empty
                if dmr.empty:
                    raise pd.errors.EmptyDataError('DMR is empty')
                # the following assumes that `fastcat` / `bamstats` has been run
                # with `-s` which is not necessarily the case (if there was only
                # one sample) --> create a dummy 'sample_name' column if missing
                if 'sample_name' not in dmr.columns:
                    dmr['sample_name'] = 'sample'
                # If one sample, save one plot
                if len(dmr['sample_name'].unique()) == 1:
                    # we only got a single sample --> no dropdown
                    EZChart(karyomap(
                        dmr, 'chrom', 'start', 'areaStat',
                        ref_lengths=faidx, stats='median'),
                        'epi2melabs')
                else:
                    # several samples --> use a dropdown menu
                    tabs = Tabs()
                    with tabs.add_dropdown_menu():
                        for sample_id, df_sample in dmr.groupby('sample_name'):
                            with tabs.add_dropdown_tab(sample_id):
                                EZChart(karyomap(
                                    df_sample, 'chrom',
                                    'start', 'areaStats',
                                    ref_lengths=faidx, stats='median'), 'epi2melabs')


# Load diff. modified sites
def load_dml(dml, faidx=None, rename=None, order=None):
    """Load dml file.

    Input is either a single directory with all the DSS DML outputs or
    a single DSS DML file.

    Optional inputs can be:
    - faidx: A faidx dataframe with 'chrom' 'length' cols (add missing intervals)
    - rename: a dictionary used to rename the sequence IDs
    - order: a ordered karyotype (dictionary of "chrom" : int(position), e.g.
    {'chr1': 1, ..., 'chrY': 24})
    """
    relevant_stats_cols_dtypes = {
        "chr": CATEGORICAL,
        "pos": int,
        "mu1": float,
        "mu2": float,
        "diff": float,
        "diff.se": float,
        "stat": float,
        "phi1": float,
        "phi2": float,
        "pval": float,
        "fdr": float,
        "postprob.overThreshold": float,
    }
    # Check inputs
    dfs = []
    if os.path.isdir(dml):
        input_files = [(dml, i) for i in os.listdir(dml)]
    elif os.path.isfile(dml):
        input_files = [(None, dml)]
    else:
        raise Exception(f'No valid input: {dml}')
    # If no files, raise error
    if len(input_files) == 0:
        raise FileNotFoundError(f'No valid input found in {dml}')
    # Process each file
    for (inpath, fname) in input_files:
        try:
            # Load the data
            dml_file = fname if not inpath else f'{inpath}/{fname}'
            tdf = pd.read_csv(dml_file, sep="\t", dtype=relevant_stats_cols_dtypes) \
                .rename(columns={'chr': 'chrom'}) \
                .eval("neg_log10_p = abs(log10(fdr))")
            if tdf.empty:
                cols = relevant_stats_cols_dtypes.update(
                    {'cum_pos': str, 'filename': int})
                dfs.append(pd.DataFrame(columns=cols))
                continue
            # Fix category names to use chrN format
            if rename:
                tdf['chrom'] = tdf['chrom'].cat.rename_categories(rename)
            if order:
                tdf = tdf \
                    .eval("chr_id = chrom.map(@order)") \
                    .sort_values(["chr_id", "pos"]) \
                    .drop(columns="chr_id")
            # Drop unused categories and reorder the remaining
            tdf['chrom'] = \
                tdf.chrom.cat.remove_unused_categories()
            tdf['chrom'] = \
                tdf.chrom.cat.reorder_categories(
                [i for i in tdf.chrom.unique()])
            # If the dataframe is empty, append it directly.
            if isinstance(faidx, pd.DataFrame):
                if 'length' not in faidx.columns:
                    raise ValueError(
                        'No "length" column found.',
                        'Import the fai with fasta_idx() and try again.')
                if 'chrom' not in faidx.columns:
                    raise ValueError(
                        'No "chrom" column found.',
                        'Import the fai with fasta_idx() and try again.')
                ref_lengths = faidx.groupby(
                    "chrom", observed=True, sort=False)["length"].last()
                total_ref_starts = ref_lengths.cumsum().shift(1, fill_value=0)
            else:
                # Compute reference lenghts from the input file
                ref_lengths = tdf.groupby(
                    "chrom", observed=True, sort=False)["pos"].last()
                # Compute cumulative position
                total_ref_starts = ref_lengths \
                    .cumsum() \
                    .shift(1, fill_value=0)
            # Compute the cumulative position.
            tdf["cum_pos"] = tdf.apply(
                lambda x: x.pos + total_ref_starts[x.chrom], axis=1)
            # Add file name
            tdf['filename'] = fname.split('/')[-1]
            # Add to output list
            dfs += [tdf
                    .reset_index(drop=True)]
        except pd.errors.EmptyDataError:
            cols = relevant_stats_cols_dtypes.update({'filename': str})
            dfs.append(pd.DataFrame(columns=cols))
    return pd.concat(dfs).reset_index(drop=True)


# Load diff. modified regions
def load_dmr(dmr, faidx=None, rename=None, order=None):
    """Load dmr file.

    Input is either a single directory with all the DSS DMR outputs or
    a single DSS DMR file.

    Optional inputs can be:
    - faidx: A faidx dataframe with 'chrom' 'length' cols (add missing intervals)
    - rename: a dictionary used to rename the sequence IDs
    - order: a ordered karyotype (dictionary of "chrom" : int(position), e.g.
    {'chr1': 1, ..., 'chrY': 24})
    """
    relevant_stats_cols_dtypes = {
        "chr": CATEGORICAL,
        "start": int,
        "end": int,
        "length": int,
        "nCG": int,
        "meanMethy1": float,
        "meanMethy2": float,
        "diff.Methy": float,
        "areaStat": float,
        }
    # Check inputs
    dfs = []
    if os.path.isdir(dmr):
        input_files = [(dmr, i) for i in os.listdir(dmr)]
    elif os.path.isfile(dmr):
        input_files = [(None, dmr)]
    else:
        raise Exception(f'No valid input: {dmr}')
    # If no files, raise error
    if len(input_files) == 0:
        raise FileNotFoundError(f'No valid input found in {dmr}')
    # Process each file
    for (inpath, fname) in input_files:
        try:
            dmr_file = fname if not inpath else f'{inpath}/{fname}'
            tdf = pd.read_csv(
                    dmr_file,
                    sep="\t",
                    dtype=relevant_stats_cols_dtypes
                    ) \
                .rename(columns={'chr': 'chrom'}) \
                .eval('abs_diff = abs(`diff.Methy`)')\
                .drop(columns=['abs_diff']) \
                .eval("mean_pos=start+(length/2)") \
                .astype({'chrom': CATEGORICAL, 'mean_pos': int})
            # If it's empty, add an empty DF
            if tdf.empty:
                cols = relevant_stats_cols_dtypes.update(
                    {'cum_pos': str, 'filename': int})
                dfs.append(pd.DataFrame(columns=cols))
                continue
            # Rename categories if required
            if rename:
                tdf['chrom'] = tdf['chrom'].cat.rename_categories(rename)
            # Order the categories
            if order:
                tdf = tdf\
                    .eval("chr_id = chrom.map(@order)") \
                    .sort_values(["chr_id", "mean_pos"]) \
                    .drop(columns="chr_id")
            # If the dataframe is empty, append it directly.
            if tdf.empty:
                dfs.append(tdf.round(4))
                continue
            # If the dataframe is empty, append it directly.
            if isinstance(faidx, pd.DataFrame):
                if 'length' not in faidx.columns:
                    raise ValueError(
                        'No "length" column found.',
                        'Import the fai with fasta_idx() and try again.')
                if 'chrom' not in faidx.columns:
                    raise ValueError(
                        'No "chrom" column found.',
                        'Import the fai with fasta_idx() and try again.')
                ref_lengths = faidx.groupby(
                    "chrom", observed=True, sort=False)["length"].last()
                total_ref_starts = ref_lengths.cumsum().shift(1, fill_value=0)
            else:
                # Compute reference lenghts from the input file
                ref_lengths = tdf.groupby(
                    "chrom", observed=True, sort=False)["end"].last()
                # Compute cumulative position
                total_ref_starts = ref_lengths \
                    .cumsum() \
                    .shift(1, fill_value=0)
            # Compute the cumulative position.
            tdf["cum_pos"] = tdf.apply(
                lambda x: x.mean_pos + total_ref_starts[x.chrom], axis=1)
            # Add file name
            tdf['filename'] = fname.split('/')[-1]
            dfs.append(tdf.round(4))
        except pd.errors.EmptyDataError:
            cols = relevant_stats_cols_dtypes.update({'filename': str})
            dfs.append(pd.DataFrame(columns=cols))
    return pd.concat(dfs).reset_index(drop=True)


def main(args):
    """Entry point to demonstrate a differentially modified component."""
    comp_title = 'DML/DMR results'
    seq_sum = DMSummary(
        dml=args.dml,
        dmr=args.dmr,
        faidx=args.faidx
        )
    report = ComponentReport(comp_title, seq_sum)
    report.write(args.output)


def argparser():
    """Argument parser for entrypoint."""
    parser = argparse.ArgumentParser(
        'DSS summary',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        add_help=False)
    parser.add_argument(
        "--dml",
        default=resource_filename('ezcharts', "data/test/test_dml.tsv.gz"),
        help="Resulting differentially modified loci from DSS.")
    parser.add_argument(
        "--dmr",
        default=resource_filename('ezcharts', "data/test/test_dmr.tsv.gz"),
        help="Resulting differentially modified regions from DSS.")
    parser.add_argument(
        "--faidx", required=False,
        default=resource_filename('ezcharts', "data/test/ref.fa.fai"),
        help="Reference fasta fai index.")
    parser.add_argument(
        "--output",
        default="dss_report.html",
        help="Output HTML file.")
    return parser



"""Test functions in utils."""

from contextlib import contextmanager
import os
import tempfile

import pandas as pd
from pkg_resources import resource_filename
import pytest

from ezcharts.components import fastcat


def _read_pandas(fname, dtype, target_cols=None):
    # this is a blind reimplementation that handles all file types
    # and simply in order to make this not just the module code

    flagstat = set(dtype.keys()) == set(fastcat.BAMSTATS_FLAGSTAT_COLS_DTYPES.keys())
    kwargs = dict()

    if not flagstat:
        kwargs["parse_dates"] = ["start_time"]
    df = pd.read_csv(
        fname, sep="\t", usecols=dtype.keys(), dtype=dtype, **kwargs
    ).reset_index(drop=True)

    if flagstat:
        df["status"] = df["ref"].apply(lambda x: "Unmapped" if x == "*" else "Mapped")
    else:
        df["start_time"] = pd.to_datetime(df.start_time, utc=True).dt.tz_localize(None)

    # homogenise fastcat and bamstats names
    df.rename(columns={"name": "read_id"}, inplace=True)

    if target_cols is not None:
        df = df[target_cols]
    return df


def _read_histograms(hist_metric, dtypes):
    hist = pd.read_csv(
        hist_metric, sep="\t", names=["start", "end", "count"], dtype=dtypes
    )
    return hist


def _compare_frames(actual, expected):
    pd.testing.assert_frame_equal(
        actual,
        expected,
        check_dtype=True,
        check_categorical=True,
        check_exact=True,
        check_datetimelike_compat=True,
    )


@contextmanager
def _empty_file(dtype=None):
    """Create an emtpy file or just with the headers."""
    with tempfile.NamedTemporaryFile(mode="r+", suffix=".tsv") as fname:
        if dtype:
            empty_df = pd.DataFrame(columns=dtype)
            empty_df.to_csv(fname.name, sep="\t")
        yield fname.name


def _multisamples():
    samples = ["S1", "S2", "S3"]
    files = [_empty_file() for i in samples]
    return samples, files


def test_001_read_fastcat():
    """Reading a fastcat file."""
    fname = resource_filename(
        "ezcharts", "data/test/real_data_test/fastcat/barcode01/per-read-stats.tsv.gz"
    )
    actual = fastcat.load_fastcat(fname)
    expected = _read_pandas(fname, fastcat.FASTCAT_COLS_DTYPES)
    _compare_frames(actual, expected)


def test_002_read_fastcat_subset():
    """Reading a fastcat with a subset of columns."""
    fname = resource_filename(
        "ezcharts", "data/test/real_data_test/fastcat/barcode01/per-read-stats.tsv.gz"
    )
    target_cols = ["sample_name", "read_length", "mean_quality"]
    for i in range(1, 3):
        actual = fastcat.load_bamstats(fname, target_cols=target_cols[:i])
        expected = _read_pandas(
            fname, fastcat.FASTCAT_COLS_DTYPES, target_cols=target_cols[:i]
        )
        _compare_frames(expected, actual)


def test_004_fastcat_just_headers():
    """Can read a file with no records."""
    with _empty_file(dtype=fastcat.FASTCAT_COLS_DTYPES) as fname:
        actual = fastcat.load_fastcat(fname).reset_index(drop=True)
        expected = _read_pandas(fname, fastcat.FASTCAT_COLS_DTYPES)
        _compare_frames(actual, expected)


def test_005_fastcat_empty():
    """Should raise an error if file completely empty."""
    with _empty_file() as fname:
        with pytest.raises(Exception, match=f"Empty input: {fname}"):
            fastcat.load_fastcat(fname)


def test_011_read_bamstats():
    """Reading a bamstats file."""
    fname = resource_filename(
        "ezcharts",
        "data/test/real_data_test/bamstats/barcode01/bamstats.readstats.tsv.gz",
    )
    actual = fastcat.load_bamstats(fname)
    expected = _read_pandas(fname, fastcat.BAMSTATS_COLS_DTYPES)
    _compare_frames(expected, actual)


def test_012_read_bamstats_subset():
    """Reading a bamstats with a subset of columns."""
    fname = resource_filename(
        "ezcharts",
        "data/test/real_data_test/bamstats/barcode01/bamstats.readstats.tsv.gz",
    )
    target_cols = ["sample_name", "read_length", "mean_quality"]
    for i in range(1, 3):
        actual = fastcat.load_bamstats(fname, target_cols=target_cols[0:i])
        expected = _read_pandas(
            fname, fastcat.BAMSTATS_COLS_DTYPES, target_cols=target_cols[:i]
        )
        _compare_frames(expected, actual)


def test_014_bamstats_just_headers():
    """Can read a file with no records."""
    with _empty_file(dtype=fastcat.BAMSTATS_COLS_DTYPES.keys()) as fname:
        actual = fastcat.load_bamstats(fname).reset_index(drop=True)
        expected = _read_pandas(fname, fastcat.BAMSTATS_COLS_DTYPES)
        _compare_frames(actual, expected)


def test_015_bamstats_empty():
    """Should raise an error if file completely empty."""
    with _empty_file() as fname:
        with pytest.raises(Exception, match=f"Empty input: {fname}"):
            fastcat.load_bamstats(fname)


def test_021_read_either():
    """Auto-detect file type and read."""
    per_read_stats_fastcat = resource_filename(
        "ezcharts", "data/test/real_data_test/fastcat/barcode01/per-read-stats.tsv.gz"
    )
    per_read_stats_bamstats = resource_filename(
        "ezcharts",
        "data/test/real_data_test/bamstats/barcode01/bamstats.readstats.tsv.gz",
    )
    target_cols_fastcat = ["sample_name", "read_length", "mean_quality"]
    target_cols_bamstats = [
        "sample_name", "coverage", "read_length", "mean_quality", "acc"]
    to_test = (
        (per_read_stats_fastcat, fastcat.FASTCAT_COLS_DTYPES, target_cols_fastcat),
        (per_read_stats_bamstats, fastcat.BAMSTATS_COLS_DTYPES, target_cols_bamstats),
    )
    for fname, TYPE, target_cols in to_test:
        actual = fastcat.load_stats(fname)
        expected = _read_pandas(fname, TYPE, target_cols=target_cols)
        _compare_frames(actual, expected)


def test_025_equality_of_result():
    """Return a consistent data structure."""
    per_read_stats_fastcat = resource_filename(
        "ezcharts", "data/test/real_data_test/fastcat/barcode01/per-read-stats.tsv.gz"
    )
    per_read_stats_bamstats = resource_filename(
        "ezcharts",
        "data/test/real_data_test/bamstats/barcode01/bamstats.readstats.tsv.gz",
    )
    target_cols = ["sample_name", "read_length", "mean_quality"]
    actual_fastcat = (
        fastcat.load_stats(per_read_stats_fastcat, target_cols=target_cols)
        .sort_values(["read_length", "mean_quality"])
        .reset_index(drop=True)
    )
    actual_bamstats = (
        fastcat.load_stats(per_read_stats_bamstats, target_cols=target_cols)
        .sort_values(["read_length", "mean_quality"])
        .reset_index(drop=True)
    )
    _compare_frames(actual_fastcat, actual_bamstats)


def test_031_read_flagstat():
    """Read a flagstats file."""
    fname = resource_filename("ezcharts", "data/test/bamstats.flagstat.tsv")
    actual = fastcat.load_bamstats_flagstat(fname)
    expected = _read_pandas(fname, fastcat.BAMSTATS_FLAGSTAT_COLS_DTYPES)
    _compare_frames(actual, expected)


def test_034_read_flagstat_just_headers():
    """Can read a file with no records."""
    with _empty_file(dtype=fastcat.BAMSTATS_FLAGSTAT_COLS_DTYPES.keys()) as fname:
        actual = fastcat.load_bamstats_flagstat(fname).reset_index(drop=True)
        expected = _read_pandas(fname, fastcat.BAMSTATS_FLAGSTAT_COLS_DTYPES)
        _compare_frames(actual, expected)


def test_035_read_flagstat_empty():
    """Should raise an error if file completely empty."""
    with _empty_file() as fname:
        with pytest.raises(Exception, match=f"Empty input: {fname}"):
            fastcat.load_bamstats_flagstat(fname)


def test_041_load_histogram():
    """Test load histogram."""
    # TODO: test histograms from bamstats
    hist_dir = resource_filename(
        "ezcharts",
        "data/test/real_data_test/bamstats/barcode01/",
    )
    variables = {
        "quality",
        "length",
        "quality.unmap",
        "length.unmap",
        "accuracy",
        "coverage",
    }
    for v in variables:
        actual = fastcat.load_histogram(hist_dir, v)
        if "length" in v:
            cols = {"start": int, "end": int, "count": int}
        else:
            cols = {"start": float, "end": float, "count": int}
        expected = _read_histograms(os.path.join(hist_dir, f"{v}.hist"), dtypes=cols)
        _compare_frames(actual, expected)


def test_042_load_unallowed_histogram():
    """Load an unallowed type."""
    hist_dir = resource_filename(
        "ezcharts",
        "data/test/real_data_test/fastcat/barcode01/",
    )
    with pytest.raises(ValueError, match=r"^`dtype` must be one of"):
        fastcat.load_histogram(hist_dir, "unallowed")


def test_101_multi_sample():
    """Raise an error if there is a discordance between number of samples and files."""
    # Create a couple of samples
    samples, files = _multisamples()
    with pytest.raises(
        ValueError,
        match="`sample_names` must be provided as a tuple "
        + "when more than one input provided.",
    ):
        # pass just first sample
        fastcat.SeqSummary(seq_summary=tuple(files), sample_names=samples[0])


def test_102_discordance_samples_files():
    """Raise an error if there is a discordance between number of samples and files."""
    # Create a couple of samples
    samples, files = _multisamples()
    with pytest.raises(
        ValueError,
        match="`sample_names` must have the same length ",
    ):
        # pass one sample less than number of files.
        fastcat.SeqSummary(seq_summary=tuple(files), sample_names=tuple(samples[:-1]))


def test_103_pass_lists():
    """Raise an error if not expected input structures are passed."""
    # Create a couple of samples
    samples, files = _multisamples()
    with pytest.raises(
        ValueError,
        match="`seq_summary` must be a path to a fastcat/bamstats ",
    ):
        # pass one sample less than number of files.
        fastcat.SeqSummary(seq_summary=files, sample_names=samples[:-1])


@pytest.mark.parametrize(
    "input",
    [
        "data/test/fastcat/f1.tsv.gz",  # file
        "data/test/histogram_stats/sample_1",  # fastcat histograms
        [  # more than 1 sample
            "data/test/fastcat/f1.tsv.gz",
            "data/test/fastcat/f2.tsv.gz",
        ],
        [  # more than 1 sample from histograms
            "data/test/histogram_stats/sample_1",
            "data/test/histogram_stats/sample_2",
        ],
    ],
)
def test_104_SeqSummary(input):
    """Create the SeqSummary component with various inputs."""
    if not isinstance(input, list):
        fastcat.SeqSummary(seq_summary=resource_filename("ezcharts", input))
    else:
        fastcat.SeqSummary(
            seq_summary=tuple([resource_filename("ezcharts", i) for i in input]),
            sample_names=tuple(["S" + i for i in input]),
        )


def test_105_bamstats_inputs():
    """Test if SeqSummary works with bam_flagstat inputs."""
    # Create a couple of samples
    seq_summary = resource_filename(
        "ezcharts",
        "data/test/real_data_test/bamstats/barcode01/bamstats.readstats.tsv.gz"
    )
    flagstats = resource_filename(
        "ezcharts",
        "data/test/real_data_test/bamstats/barcode01/bamstats.flagstat.tsv"
    )
    # pass just first sample
    fastcat.SeqSummary(
        seq_summary=seq_summary,
        sample_names="barcode01",
        flagstat=flagstats)


def test_106_multi_sample_by_metric():
    """Test if SeqCompare works."""
    # Create a couple of samples
    samples, files = _multisamples()
    with pytest.raises(
        ValueError,
        match="`sample_names` must be provided as a tuple "
        + "when more than one input provided.",
    ):
        # pass just first sample
        fastcat.SeqCompare(
            seq_summary=tuple(files), sample_names=samples[0])


def test_107_SeqSummary_empty_file():
    """Create the SeqSummary component with an empty file (summaries)."""
    with _empty_file(dtype=fastcat.FASTCAT_COLS_DTYPES) as fname:
        fastcat.SeqSummary(seq_summary=fname)


def test_108_SeqSummary_empty_histogram_files():
    """Create the SeqSummary component with an histograms dir with empty files."""
    fastcat.SeqSummary(
        seq_summary=resource_filename(
            "ezcharts", "data/test/histogram_stats/empty_sample/"
        )
    )



"""Get ezchart containers."""
from abc import ABC, abstractmethod
import json

from dominate.tags import script
from dominate.util import raw

from ezcharts.layout.base import Snippet
from ezcharts.layout.util import render_template
from ezcharts.plots import BokehPlot, Plot


def EZChart(
    plot,
    theme: str = 'epi2melabs',
    width: str = '100%',
    height: str = '500px',
    additional_styles: dict = None
):
    """Wrap an ECharts or Bokeh plot in a div."""
    if isinstance(plot, BokehPlot):
        return _BokehChart(
            plot, theme=theme, width=width, height=height,
            additional_styles=additional_styles)
    elif isinstance(plot, Plot):
        return _EChart(
            plot, theme=theme, width=width, height=height,
            additional_styles=additional_styles)
    else:
        raise ValueError(f"`EZChart()` called with argument of unexpected type: {plot}")


class _ReportChart(ABC, Snippet):
    """ABC for `div` containers containing ECharts or Bokeh plots."""

    TAG = 'div'

    @abstractmethod
    def __init__(self, width, height, class_name, additional_styles=None):
        """Set width and height of plot."""
        style_dict = {
            "width": width,
            "height": height,
            **(additional_styles or {})
        }

        style_str = "; ".join(
            f"{key}:{value}" for key, value in style_dict.items())

        Snippet.__init__(
            self,
            styles=None,
            classes=None,
            className=class_name,
            style=style_str)


class _BokehChart(_ReportChart):
    """Wraps a Bokeh plot in a div."""

    def __init__(
        self,
        plot,
        theme: str = 'epi2melabs',
        width: str = '100%',
        height: str = '500px',
        additional_styles=None
    ) -> None:
        """Add placeholder div for BokehChart.

        The JS for all Bokeh plots will be added at report generation; no need to do
        anything besides keeping a ref to the plot at this point.
        """
        super().__init__(
            width, height, class_name="bokeh-chart-container",
            additional_styles=additional_styles)
        self.plot = plot


class _EChart(_ReportChart):
    """Wraps an ECharts plot in a div."""

    def __init__(
        self,
        plot,
        theme: str = 'epi2melabs',
        width: str = '100%',
        height: str = '500px',
        additional_styles=None
    ) -> None:
        """Create a div and script tag for initialising the plot."""
        # `class_name="echarts-chart-container"` here is required for resizing the chart
        # when a different tab is selected (c.f. `update_charts_on_tab_change()` in
        # `data/scripts/epi2melabs.js`)
        super().__init__(
            width, height, class_name="echarts-chart-container",
            additional_styles=additional_styles)

        with self:
            script(raw(render_template(
                """
                var chart_{{ id }} = echarts.init(
                    dom=document.getElementById('{{ id }}'),
                    theme='{{ t }}',
                    opts={renderer: '{{ c.renderer }}'});
                var opt_{{ id }} = {{ j | replace('"',"'") | safe }};
                chart_{{ id }}.setOption(opt_{{ id }});
                {% if w.endswith('%') %}
                    window.addEventListener('resize', function(){
                        chart_{{ id }}.resize();
                    })
                {% endif %}
                chart_{{ id }}.resize()
                """,
                c=plot, j=plot.to_json(), t=theme,
                w=width, id=self.uid)))


class EZChartTheme(script):
    """Loads an echarts theme in a script tag."""

    def __init__(
        self,
        theme: str
    ) -> None:
        """Load theme."""
        super().__init__(tagname='script')
        with self:
            raw(render_template(
                """
                echarts.registerTheme(
                    '{{ n.themeName }}',
                    {{ t | replace(\'"\',"\'") | safe }}.theme);
                """,
                n=theme, t=json.dumps(theme)))


# No module demo here as entrypoint is in `plots.demo`.



"""An ezcharts component for plotting sequence summaries."""
import argparse
import copy
import os

from bokeh.models import Title
import numpy as np
import pandas as pd
from pkg_resources import resource_filename
import sigfig

import ezcharts as ezc
from ezcharts.components.ezchart import EZChart
from ezcharts.components.reports.comp import ComponentReport
from ezcharts.layout.base import Snippet
from ezcharts.layout.snippets import DataTable, Grid, Tabs
from ezcharts.plots import BokehPlot


FASTCAT_COLS_DTYPES = {
    "read_id": str,
    "filename": "category",
    "sample_name": "category",
    "read_length": int,
    "mean_quality": float,
    "channel": int,
    "read_number": int,
    "start_time": str,
}

BAMSTATS_COLS_DTYPES = {
    "name": str,
    "sample_name": "category",
    "ref": "category",
    "coverage": float,
    "ref_coverage": float,
    "qstart": "Int64",
    "qend": "Int64",
    "rstart": "Int64",
    "rend": "Int64",
    "aligned_ref_len": "Int64",
    "direction": "category",
    "length": int,
    "read_length": int,
    "mean_quality": float,
    "start_time": str,
    "match": int,
    "ins": int,
    "del": int,
    "sub": int,
    "iden": float,
    "acc": float,
    "duplex": int,
}

BAMSTATS_FLAGSTAT_COLS_DTYPES = {
    "ref": "category",
    "sample_name": "category",
    "total": int,
    "primary": int,
    "secondary": int,
    "supplementary": int,
    "unmapped": int,
    "qcfail": int,
    "duplicate": int,
    "duplex": int,
    "duplex_forming": int,
}

TIMES = ["start_time"]


def _intersect_columns(base, requested=None):
    # Take one of the dtype dictionaries and intersect with a list
    cols = base
    if requested is not None:
        if invalid_cols := set(requested) - set(base.keys()):
            raise ValueError(f"Found unexpected columns {invalid_cols}.")
        cols = {col: base[col] for col in requested}
    times = [t for t in TIMES if t in cols]
    if len(times) == 0:
        times = None
    return cols, times


def _localize_time(df):
    # ensure we have datetime format, convert to utc
    for t in TIMES:
        if t in df.columns:
            df[t] = pd.to_datetime(df.start_time, utc=True).dt.tz_localize(None)
    return df


class SeqSummary(Snippet):
    """Generate sequence summary plots."""

    def __init__(
        self,
        seq_summary=None,
        flagstat=None,
        sample_names=None,
        theme="epi2melabs",
        color=None,
        height="500px",
        alignment_stats=True,
        # [CW-5211]
        # When users are sequencing small molecules they sometimes find
        # spurious "long" reads which are either real concatemers or artefacts
        # these have the effect of causing aggressive binning in the read length
        # plot such that any detail is lost from the main distribution.
        # This arg allows the binwidth of the length plot to be customised to
        # customer requirements.
        read_length_plot_binwidth=None
    ):
        """Create sequence summary component.

        :param seq_summary: a path to a fastcat/bamstats read stats output file
            or dataframe (or a tuple of such).
        :param flagstat: a path to a bamstats flagstat output file or dataframe
            (or tuple of such).
        :sample_names: tuple of sample names. Required when other input arguments
            are tuples.
        :param theme: String defining the visual theme for the plots.
        :param color: String defining the color for the plots.
        :param height: String defining the height of the plots.
        :param alignment_stats: Boolean defining whether to display the alignments
         stats.
        """
        super().__init__(styles=None, classes=None)
        self.theme = theme
        self.color = color
        self.read_length_plot_binwidth = read_length_plot_binwidth

        # we need at least seq_summary or histograms
        if seq_summary is None:
            raise ValueError("One of `seq_summary` must be provided.")
        if not (
            isinstance(seq_summary, pd.DataFrame)
            or isinstance(seq_summary, str)
            or isinstance(seq_summary, tuple)
        ):
            raise ValueError(
                "`seq_summary` must be a path to a fastcat/bamstats read stats output "
                "file or dataframe (or a tuple of such)."
            )

        # check sample_names is a tuple if files are provided in tuples.
        if (isinstance(seq_summary, tuple)) and not isinstance(sample_names, tuple):
            raise ValueError(
                "`sample_names` must be provided as a tuple "
                "when more than one input provided.")

        # check tuples are all the same length
        if isinstance(sample_names, tuple) and (len(sample_names) != len(seq_summary)):
            raise ValueError(
                "`sample_names` must have the same length as `seq_summary`.")
        if (
                (flagstat is not None)
                and isinstance(flagstat, tuple)
                and (len(sample_names) != len(flagstat))):
            raise ValueError(
                "`sample_names` must have the same length as `flagstat`.")
        with self:
            if isinstance(seq_summary, tuple):
                # several samples => use a dropdown
                tabs = Tabs()
                with tabs.add_dropdown_menu():
                    for sample_name, data in sorted(
                        zip(sample_names, seq_summary), key=lambda x: x[0]
                    ):
                        with tabs.add_dropdown_tab(sample_name):
                            ldata, qdata, adata, cdata = self._load_summary_data(data)
                            self._draw_summary_plots(ldata, qdata, height)
                            if alignment_stats:
                                self._draw_alignment_plots(adata, cdata, height)
            else:
                # single sample
                ldata, qdata, adata, cdata = self._load_summary_data(seq_summary)
                self._draw_summary_plots(ldata, qdata, height)
                if alignment_stats:
                    self._draw_alignment_plots(adata, cdata, height)

            # same again for flagstat
            if flagstat is not None:
                if isinstance(flagstat, tuple):
                    tabs = Tabs()
                    with tabs.add_dropdown_menu():
                        for sample_name, data in sorted(
                            zip(sample_names, flagstat), key=lambda x: x[0]
                        ):
                            with tabs.add_dropdown_tab(sample_name):
                                self._draw_bamstat_table(data)
                else:
                    self._draw_bamstat_table(flagstat)

    def _load_summary_data(
        self,
        data
    ):
        """Draw quality, read_length, yield, accuracy and coverage plots using raw data.

        :param seq_summary: pd.DataFrame containing per-sequence summary information.
        :param theme: EPI2ME theme.
        """
        # data is either a summary file, a dataframe (of a summary file),
        # or a directory (of histogram files).
        adata, cdata, qdata, ldata = None, None, None, None
        if isinstance(data, pd.DataFrame):
            qdata, ldata = data, data
            adata = data if 'acc' in data.columns else None
            cdata = data if 'coverage' in data.columns else None
        else:
            try:
                df = load_stats(data)
                qdata, ldata = df, df
                adata = df if 'acc' in df.columns else None
                cdata = df if 'coverage' in df.columns else None
            except Exception:
                try:
                    qdata = load_histogram(data, "quality")
                    ldata = load_histogram(data, "length")
                    # Try loading BAMstats files
                    try:
                        adata = load_histogram(data, "accuracy")
                        cdata = load_histogram(data, "coverage")
                    except Exception:
                        adata, cdata = None, None
                    # Try loading unmapped files
                    try:
                        qdata = sum_hists((
                            qdata,
                            load_histogram(data, "quality.unmap")
                        ))
                        ldata = sum_hists((
                            ldata,
                            load_histogram(data, "length.unmap")
                        ))
                    except Exception:
                        qdata, ldata

                except Exception:
                    raise ValueError("Could not load input data.")
        return ldata, qdata, adata, cdata

    def _draw_summary_plots(
        self,
        ldata,
        qdata,
        height,
    ):
        """Draw quality, read_length and yield plots using raw data.

        :param ldata: pd.DataFrame containing length values.
        :param qdata: pd.DataFrame containing quality values.
        :param color: color for the plots.
        :param height: height of the plot
        """
        with Grid(columns=3):
            EZChart(
                read_quality_plot(qdata, color=self.color), self.theme, height=height)
            EZChart(
                read_length_plot(
                    ldata, color=self.color, binwidth=self.read_length_plot_binwidth),
                self.theme, height=height)
            EZChart(
                base_yield_plot(ldata, color=self.color), self.theme, height=height)

    def _draw_alignment_plots(
        self,
        adata,
        cdata,
        height,
    ):
        """Draw accuracy and coverage plots using raw data.

        :param adata: pd.DataFrame containing accuracy values.
        :param cdata: pd.DataFrame containing coverage values.
        """
        if adata is not None or cdata is not None:
            with Grid(columns=2):
                if adata is not None:
                    try:
                        EZChart(
                            mapping_accuracy_plot(adata, color=self.color),
                            self.theme, height=height)
                    except Exception:
                        pass
                if cdata is not None:
                    try:
                        EZChart(
                            read_coverage_plot(cdata, color=self.color),
                            self.theme, height=height)
                    except Exception:
                        pass

    def _draw_bamstat_table(self, data):
        if not isinstance(data, pd.DataFrame):
            data = load_bamstats_flagstat(data)
        DataTable.from_pandas(data, use_index=False)


class SeqCompare(Snippet):
    """Generate sequence comparison plots."""

    def __init__(
        self,
        seq_summary=None,
        flagstat=None,
        sample_names=None,
        theme="epi2melabs",
        color=None,
        height="500px",
        alignment_stats=True,
    ):
        """Create sequence summary component.

        :param seq_summary: a path to a fastcat/bamstats read stats output file
            or dataframe (or a tuple of such).
        :param flagstat: a path to a bamstats flagstat output file or dataframe
            (or tuple of such).
        :sample_names: tuple of sample names. Required when other input arguments
            are tuples.
        :param theme: String defining the visual theme for the plots.
        """
        super().__init__(styles=None, classes=None)
        self.theme = theme
        self.color = color
        self.alignment_stats = alignment_stats
        self.metrics = [
            'length',
            'yield',
            'quality',
            'accuracy',
            'coverage'
        ]
        self.columns = [
            'read_length',
            'read_length',
            'mean_quality',
            'acc',
            'coverage'
        ]
        self.plot_functions = [
            read_length_plot,
            base_yield_plot,
            read_quality_plot,
            mapping_accuracy_plot,
            read_coverage_plot
        ]

        # we need at least seq_summary or histograms
        if seq_summary is None:
            raise ValueError("One of `seq_summary` must be provided.")
        if not (
            isinstance(seq_summary, pd.DataFrame)
            or isinstance(seq_summary, str)
            or isinstance(seq_summary, tuple)
        ):
            raise ValueError(
                "`seq_summary` must be a path to a fastcat/bamstats read stats output "
                "file or dataframe (or a tuple of such)."
            )

        # check sample_names is a tuple if files are provided in tuples.
        if (isinstance(seq_summary, tuple)) and not isinstance(sample_names, tuple):
            raise ValueError(
                "`sample_names` must be provided as a tuple "
                "when more than one input provided.")

        # check tuples are all the same length
        if isinstance(sample_names, tuple) and (len(sample_names) != len(seq_summary)):
            raise ValueError(
                "`sample_names` must have the same length as `seq_summary`.")
        if (
                (flagstat is not None)
                and isinstance(flagstat, tuple)
                and (len(sample_names) != len(flagstat))):
            raise ValueError(
                "`sample_names` must have the same length as `flagstat`.")

        with self:
            if isinstance(seq_summary, tuple):
                self._compare_summary_plots(
                    sample_names, seq_summary, height=height)
            else:
                # single sample
                self._compare_summary_plots(
                    tuple('Sample'), tuple(seq_summary), height=height)

            # same again for flagstat
            if flagstat is not None:
                if isinstance(flagstat, tuple):
                    tabs = Tabs()
                    with tabs.add_dropdown_menu():
                        for sample_name, data in sorted(
                            zip(sample_names, flagstat), key=lambda x: x[0]
                        ):
                            with tabs.add_dropdown_tab(sample_name):
                                self._draw_bamstat_table(data)
                else:
                    self._draw_bamstat_table(flagstat)

    def _compare_summary_plots(self, samples, datasets, height):
        """Draw quality, read_length, yield, accuracy and coverage plots using raw data.

        :param samples: tuple of sample names to process.
        :param datasets: data to process, in the same order as the samples.
        """
        # data is either a summary file, a dataframe (of a summary file),
        # or a directory (of histogram files).
        qdata, ldata = [], []
        all_plots = {
            'length': [],
            'yield': [],
            'quality': [],
            'accuracy': [],
            'coverage': []

        }
        # For each metric and each sample, create the plot
        if not isinstance(datasets, tuple):
            raise ValueError(
                "SeqSummary by `metric` only Tuples as input."
            )
        else:
            # TODO: test with one DF directly
            if isinstance(datasets, pd.DataFrame):
                for (metric, col, plot_function) in zip(
                        self.metrics, self.columns, self.plot_functions):
                    if col in datasets.columns:
                        all_plots[metric] += [plot_function(datasets, color=self.color)]
                    else:
                        all_plots[metric] += [None]
            for (sample, data) in zip(samples, datasets):
                try:
                    df = load_stats(data)
                    for (metric, col, plot_function) in zip(
                            self.metrics, self.columns, self.plot_functions):
                        if col in df.columns:
                            all_plots[metric] += [plot_function(df, color=self.color)]
                        else:
                            all_plots[metric] += [None]

                except Exception:
                    # Bare minimum, should always be there
                    try:
                        qdata = load_histogram(data, "quality")
                        ldata = load_histogram(data, "length")
                    except Exception:
                        raise ValueError("Could not load input data.")
                    # Try loading and adding unmapped hists, if available
                    try:
                        qdata = sum_hists((
                            load_histogram(data, "quality"),
                            load_histogram(data, "quality.unmap")
                        ))
                        ldata = sum_hists((
                            load_histogram(data, "length"),
                            load_histogram(data, "length.unmap")
                        ))
                    except Exception:
                        qdata, ldata
                    all_plots['length'] += [read_length_plot(ldata, color=self.color)]
                    all_plots['yield'] += [base_yield_plot(ldata, color=self.color)]
                    all_plots['quality'] += [read_quality_plot(qdata, color=self.color)]
                    # Try loading BAMstats-specific hists and add their plots, if needed
                    if self.alignment_stats:
                        try:
                            all_plots['accuracy'] += [
                                mapping_accuracy_plot(
                                    load_histogram(data, "accuracy"),
                                    color=self.color
                                )
                            ]
                        except Exception:
                            all_plots['accuracy'] += [None]
                        try:
                            all_plots['coverage'] += [
                                read_coverage_plot(
                                    load_histogram(data, "coverage"),
                                    color=self.color
                                )
                            ]
                        except Exception:
                            all_plots['coverage'] += [None]

        # Create tabs
        tabs = Tabs()
        for metric in self.metrics:
            pairs = [(p, s) for p, s in zip(all_plots[metric], samples) if p]
            if len(pairs) == 0:
                continue
            with tabs.add_tab(metric):
                with Grid(columns=min(len(pairs), 3)):
                    plots, samps = zip(*sorted(pairs, key=lambda x: x[1]))
                    self._coordinate_plots(plots, labels=samps)
                    for plot in plots:
                        EZChart(plot, self.theme, height=height)

    def _is_empty_plot(self, plot):
        """Check if the plot is empty."""
        if not plot:
            return True
        if isinstance(plot, BokehPlot):
            if "_fig" not in dir(plot):
                return True
            for renderer in plot._fig.renderers:
                if not hasattr(renderer, 'glyph'):
                    return True
        else:
            if not plot.dataset:
                return True
        return False

    def _coordinate_plots(self, plots, labels=None):
        """Coordinate the axes between the plots."""
        max_by_plot = []
        for plot in plots:
            if not self._is_empty_plot(plot):
                if isinstance(plot, BokehPlot):
                    dimensions = plot._fig.renderers[0].data_source.to_df()
                    if 'top' in dimensions.columns:
                        col = 'top'
                    else:
                        col = 'y'
                    y_val = dimensions[col].max()
                else:
                    dimensions = np.array(plot.dataset[0].dimensions)
                    if 'y' in dimensions:
                        y_col = np.where(dimensions == 'y')[0]
                    else:
                        y_col = np.where(dimensions == 'height')[0]
                    y_val = np.max(plot.dataset[0].source, axis=0)[y_col]
                max_by_plot.append(y_val)
        # Set new values
        if labels:
            inputs = zip(plots, labels)
        else:
            inputs = zip(plots, [None] * len(plots))
        for (plot, label) in inputs:
            if not self._is_empty_plot(plot):
                plot._fig.y_range.end = max(max_by_plot)
                if label:
                    plot._fig.add_layout(
                        Title(text=label, text_font_size="1.5em"), 'above')
            else:
                if label:
                    plot.title.text = label

    def _draw_bamstat_table(self, data):
        if not isinstance(data, pd.DataFrame):
            data = load_bamstats_flagstat(data)
        DataTable.from_pandas(data, use_index=False)


@ezc.plots.util.plot_wrapper
def base_yield_plot(data, color=None):
    """Create yield plot by plotting total yield above read length.

    :param data: fastcat/bamstats summary data or read-length histogram data.
    """
    xlab = "Read length / kb"
    ylab = "Yield above length / Gbases"
    thinning = None  # don't really know why this is different
    # note: we want an anticumulative sum, hence all the [::-1]
    if "read_length" in data.columns:
        # need to create the data
        thinning = 1000
        length = np.concatenate(([0], np.sort(data["read_length"])), dtype="int")
        cumsum = np.cumsum(length[::-1])[::-1]
    else:
        # assume histogram
        # TODO: this assumes even bins
        thinning = 100
        length = data["start"]
        cumsum = np.cumsum(
            data["start"][::-1].to_numpy() * data["count"][::-1].to_numpy()
        )[::-1]

    mid = cumsum[-1] / 2
    n50_index = np.searchsorted(cumsum, mid)
    n50 = int(length[n50_index])
    # TODO: we needn't create this only to thin it immediately
    df = pd.DataFrame({xlab: length / 1000, ylab: cumsum / 1e9}, copy=False)

    # thin the data while keeping last
    if len(df) > thinning:
        step = len(df) // thinning
        df = pd.concat((df.loc[::step, :], df.iloc[[-1]]), axis=0)

    plt = ezc.lineplot(data=df, x=xlab, y=ylab, hue=None, color=color, marker=False)
    subtext = (
        f"Total yield: {sigfig.round(df.iloc[0][ylab], sigfigs=3)} Gb. "
        f" N50: {sigfig.round(n50, 3)}kb")

    plt._fig.add_layout(Title(text=subtext, text_font_size="0.8em"), 'above')
    plt._fig.add_layout(
        Title(text="Base yield above read length", text_font_size="1.5em"), 'above')
    return plt


@ezc.plots.util.plot_wrapper
def histogram_plot(
    data, col="count", binwidth=None, min_val=None,
    max_val=None, title="Histogram",
    xaxis_label=None, color=None
):
    """Create histogram summary plot.

    :param data: fastcat/bamstats summary data or read-length histogram data.
    :param binwidth: width of each bin.
    :param min_val: the minimum value to plot.
    :param max_val: the maximum value to plot.
    :param title: title of the plot.
    """
    plt, mean_val, median_val = None, None, None
    if "read_length" in data.columns:
        if not min_val:
            min_val = data[col].min()
        if not max_val:
            max_val = data[col].max()
        # fastcat/bamstats
        mean_val = np.round(data[col].mean(), 1)
        median_val = int(data[col].median())
        plt = ezc.histplot(
            data=data[col], binwidth=binwidth, binrange=(min_val, max_val), color=color
        )
    else:
        if not min_val:
            min_val = data.end.min()
        if not max_val:
            max_val = data.end.max()
        # histogram
        mean_val = np.round(
            np.average(0.5 * (data["start"] + data["end"]), weights=data["count"]), 1
        )
        median_val = int(histogram_median(data))
        plt = ezc.histplot(
            data=data["start"],
            weights=data["count"],
            binwidth=binwidth,
            binrange=(min_val, max_val),
            color=color
        )

    subtitle = f"Mean: {mean_val:.1f}. Median: {median_val:.1f}"

    if xaxis_label:
        plt._fig.xaxis.axis_label = xaxis_label
    plt._fig.add_layout(Title(text=subtitle, text_font_size="0.8em"), 'above')
    plt._fig.add_layout(Title(text=title, text_font_size="1.5em"), 'above')
    plt._fig.x_range.start, plt._fig.x_range.end = min_val, max_val
    plt._fig.yaxis.axis_label = "Number of reads"
    return plt


@ezc.plots.util.plot_wrapper
def read_quality_plot(data, binwidth=0.2, min_qual=4, max_qual=30, color=None):
    """Create read quality summary plot.

    :param data: fastcat/bamstats summary data or read-length histogram data.
    :param binwidth: width of each bin.
    :param min_qual: the minimum quality value to plot.
    :param max_qual: the maximum quality value to plot.
    """
    plt = histogram_plot(
        data, col='mean_quality', binwidth=binwidth, min_val=min_qual,
        max_val=max_qual, title="Read quality",
        xaxis_label="Quality", color=color
    )
    return plt


@ezc.plots.util.plot_wrapper
def mapping_accuracy_plot(
    data, binwidth=1.0, min_acc=80,
    max_acc=101, color=None
):
    """Create read quality summary plot.

    :param data: fastcat/bamstats summary data or read-length histogram data.
    :param binwidth: width of each bin.
    :param min_acc: the minimum quality value to plot.
    :param max_acc: the maximum quality value to plot.
    """
    plt = histogram_plot(
        data, col='acc', binwidth=0.2, min_val=min_acc,
        max_val=max_acc, title="Accuracy",
        xaxis_label="Accuracy", color=color
    )
    return plt


@ezc.plots.util.plot_wrapper
def read_coverage_plot(
    data, binwidth=1.0, min_cov=0,
    max_cov=None, color=None
):
    """Create read quality summary plot.

    :param data: fastcat/bamstats summary data or read-length histogram data.
    :param binwidth: width of each bin.
    :param min_cov: the minimum coverage value to plot.
    :param max_cov: the maximum coverage value to plot.
    """
    plt = histogram_plot(
        data, col='coverage', binwidth=0.2, min_val=min_cov,
        max_val=max_cov, title="Read alignment",
        xaxis_label="Percentage of read aligned", color=color
    )
    return plt


@ezc.plots.util.plot_wrapper
def read_length_plot(
    data, xlim=(0, None), quantile_limits=False,
    bins=100, binwidth=None, title="Read length", color=None
):
    """Create a read length plot.

    :param seq_summary: pd.DataFrame containing per-sequence summary information.
    :param xlim: viewable read length limits.
    :param quantile_limits: if True, xlim is interpreted as quantiles of the data rather
        than absolute values.
    :param bins: number of bins.
    :param binwidth: bin width.
    :param title: plot title.

    The reads will be filtered with `min_len` and `max_len` before calculating the
    histogram. The subtext of the plot title will still show the mean / median / maximum
    of the full data.
    """
    plt, mean_length, median_length = None, None, None
    min_len, max_len = xlim
    if min_len is None:
        min_len = 0
    # create data to plot depending on input type.
    if "read_length" in data.columns:
        # fastcat/bamstats
        mean_length = np.round(data.read_length.mean(), 1)
        median_length = int(data.read_length.median())
        max_ = int(np.max(data["read_length"]))
        min_ = int(np.min(data["read_length"]))
        read_lengths = data["read_length"].values

        if max_len is None:
            max_len = 1 if quantile_limits else read_lengths.max()

        if quantile_limits:
            min_len, max_len = np.quantile(read_lengths, [min_len, max_len])
            # set `xlim` so that we can use it to set the x-axis limits below
            xlim = (min_len, max_len)

        read_lengths = read_lengths[
            (read_lengths >= min_len) & (read_lengths <= max_len)
        ]
        weights = None
    else:
        # histogram
        if max_len is None:
            max_len = 1 if quantile_limits else data.iloc[-1]["end"]
        if quantile_limits:
            cs = np.cumsum(data["count"])
            lower_idx, upper_idx = np.searchsorted(cs, np.quantile(cs, quantile_limits))
            min_len, max_len = (data["start"][lower_idx], data["end"][upper_idx])
            xlim = (min_len, max_len)

        median_length = histogram_median(data)
        mean_length = np.average(data["start"], weights=data["count"])
        max_ = data.iloc[-1]["start"]
        min_ = data.iloc[0]["start"]
        data = data[(data["start"] >= min_len) & (data["end"] <= max_len)]
        weights = data["count"]
        read_lengths = data['start']

    if binwidth is not None:
        binwidth /= 1000
    read_lengths = read_lengths / 1000
    plt = ezc.histplot(
        read_lengths,
        bins=bins,
        binwidth=binwidth,
        weights=weights,
        color=color
    )
    # customize the plot
    plt._fig.xaxis.axis_label = "Read length / kb"
    plt._fig.yaxis.axis_label = "Number of reads"
    subtext = (
        f"Mean: {mean_length:.1f}. Median: {median_length:.1f}. "
        f"Min: {min_:,d}. Max: {max_:,d}"
    )
    plt._fig.add_layout(Title(text=subtext, text_font_size="0.8em"), 'above')
    plt._fig.add_layout(Title(text=title, text_font_size="1.5em"), 'above')

    if xlim[0] is not None:
        plt._fig.x_range.start = xlim[0] / 1000
    if xlim[1] is not None:
        plt._fig.x_range.end = xlim[1] / 1000
    return plt


def sum_hists(hists):
    """Sum two histogram dataframes based on the intervals."""
    if not isinstance(hists, tuple):
        raise ValueError("Input is not a tuple of dataframes.")
    return pd.concat(
        hists
    )\
        .reset_index(drop=True)\
        .sort_values(by='start')\
        .groupby(['start', 'end'])\
        .sum().reset_index()


def load_fastcat(fpath, target_cols=None):
    """Load and prepare fastcat per-read stats.

    :param fpath: path to a fastcat stats file.
    :target_cols: columns to be loaded in the dataframe.

    :returns: a dataframe
    """
    cols, time_cols = _intersect_columns(copy.copy(FASTCAT_COLS_DTYPES), target_cols)
    try:
        df = pd.read_csv(
            fpath, sep="\t", usecols=cols.keys(), dtype=cols, parse_dates=time_cols
        )
    except pd.errors.EmptyDataError:
        raise Exception(f"Empty input: {fpath}")
    df = _localize_time(df)
    return df


def load_bamstats(fpath, target_cols=None):
    """Load and prepare bamstats per-read stats.

    :param fpath: path to a bamstats stats file.
    :target_cols: columns to be loaded in the dataframe.

    :returns: a dataframe
    """
    cols, time_cols = _intersect_columns(copy.copy(BAMSTATS_COLS_DTYPES), target_cols)
    try:
        df = pd.read_csv(
            fpath, sep="\t", usecols=cols.keys(), dtype=cols, parse_dates=time_cols
        )
    except pd.errors.EmptyDataError:
        raise Exception(f"Empty input: {fpath}")
    df = _localize_time(df)
    # to be consistent with fastcat
    df.rename(columns={"name": "read_id"}, inplace=True)
    return df


def load_bamstats_flagstat(fpath):
    """Load and prepare bamstats flagstat output.

    :param fpath: path to a bamstats flagstat file.

    :returns: a dataframe
    """
    try:
        df = pd.read_csv(
            fpath,
            sep="\t",
            usecols=BAMSTATS_FLAGSTAT_COLS_DTYPES.keys(),
            dtype=BAMSTATS_FLAGSTAT_COLS_DTYPES,
        )
    except pd.errors.EmptyDataError:
        raise Exception(f"Empty input: {fpath}")
    df["status"] = df["ref"].apply(lambda x: "Unmapped" if x == "*" else "Mapped")
    return df


def load_histogram(hist_dir, dtype="quality"):
    """Load fastcat/bamstats histograms.

    :param hist_dir: pathname to input directory.
    :param dtype: histogram datatype to load.
    """
    allowed = {
        "quality",
        "length",
        "quality.unmap",
        "length.unmap",
        "accuracy",
        "coverage",
    }
    if dtype not in allowed:
        raise ValueError(f"`dtype` must be one of {allowed}.")
    dt = float
    if "length" in dtype:
        dt = int

    hist = pd.read_csv(
        os.path.join(hist_dir, f"{dtype}.hist"),
        sep="\t",
        names=["start", "end", "count"],
        dtype={"start": dt, "end": dt, "count": int},
    )
    return hist


def load_stats(fpath, target_cols=None):
    """Load and prepare fastcat or bamstats per-read stats.

    This function is intended to be used when the caller does not know (and care)
    the origin of the input file, but wishes to load the common columns.

    :params fpath: path to a fastcat or bamstats stats file.
    :param target_cols: list of columns to read from file.

    :returns: a dataframe
    """
    target_cols_fastcat, target_cols_bamstats = target_cols, target_cols
    if target_cols is None:
        target_cols_fastcat = ["sample_name", "read_length", "mean_quality"]
        target_cols_bamstats = [
            "sample_name", "read_length", "mean_quality", "acc", "coverage"]
    df = None
    try:
        df = load_bamstats(fpath, target_cols=target_cols_bamstats)
    except ValueError:
        df = load_fastcat(fpath, target_cols=target_cols_fastcat)
    return df


def histogram_median(hist, x="start", y="count"):
    """Calculate the median value from histogram data.

    :param hist: pd.DataFrame with columns [start, end, count].
    :param x: the x-axis variable.
    :param y: the y-axis variable.
    """
    # TODO: start and end may not be end = start + 1
    cumsum = np.cumsum(hist[y]).to_numpy()
    mid = cumsum[-1] / 2
    pos = np.searchsorted(cumsum, mid)
    return hist[x][pos]


def main(args):
    """Entry point to demonstrate a sequence summary component."""
    comp_title = "Sequence Summary"
    # Define inputs.
    # If seq_summary is not passed, then use bam_readstats
    if args.seq_summary:
        seq_summary = args.seq_summary
    elif args.seq_summary is None and args.bam_readstats is not None:
        seq_summary = args.bam_readstats
    else:
        raise ValueError('No valid input for seq_summary/bam_readstats.')
    bam_flagstat = tuple(args.bam_flagstat) \
        if isinstance(args.bam_flagstat, list) \
        else args.bam_flagstat
    sample = tuple(args.sample) \
        if isinstance(args.sample, list) \
        else args.sample
    # Create summary
    if args.by == 'sample':
        seq_sum = SeqSummary(
            tuple(seq_summary),
            flagstat=bam_flagstat,
            sample_names=sample,
            alignment_stats=False if args.skip_alignment_stats else True,
            color=args.color,
        )
    else:
        seq_sum = SeqCompare(
            tuple(seq_summary),
            flagstat=bam_flagstat,
            sample_names=sample,
            alignment_stats=False if args.skip_alignment_stats else True,
            color=args.color,
        )

    # Write report
    report = ComponentReport(comp_title, seq_sum)
    report.write(args.output)


def argparser():
    """Argument parser for entrypoint."""
    parser = argparse.ArgumentParser(
        "Sequence summary",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        add_help=False,
    )
    parser.add_argument(
        "--seq_summary",
        default=resource_filename("ezcharts", "data/test/fastcat"),
        help="Sequence summary TSV from fastcat.",
        nargs='+'
    )
    parser.add_argument(
        "--bam_flagstat",
        default=resource_filename(
            "ezcharts", "data/test/bamstats/bamstats.flagstat.tsv"
        ),
        help="Bam flagstats TSV from bamstats.",
        nargs='+'
    )
    parser.add_argument(
        "--sample",
        default='sample',
        help="Sample name.",
        nargs='+'
    )
    parser.add_argument(
        "--by",
        default='sample',
        choices=['sample', 'metric'],
        help="Create SeqSummary by sample or by metric.",
    )
    parser.add_argument(
        "--bam_readstats",
        default=resource_filename("ezcharts", "data/test/bamstats/"),
        help="Read statistics TSV from bamstats.",
        nargs='+'
    )
    parser.add_argument(
        "--skip_alignment_stats",
        action="store_true",
        help="Skip alignment statistics histograms (coverage and accuracy)."
    )
    parser.add_argument(
        "--color",
        help="Plot color."
    )
    parser.add_argument(
        "--output", default="seq_summary_report.html", help="Output HTML file."
    )
    return parser
